/*
 * AlgorithmKF.java v6.1 
 * Created by: Ryan Irwin 05/24/2005
 * Modified by: Sanjay Patil 01/10/2006
 *
 * Description: Kalman Filtering

 * Reference: 
 * G. Welch and G. Bishop, "An Introduction to the Kalman filter,"
 * Technical Report TR 95-041, Department of Computer Science,
 * University of North Carolina at Chapel Hill, Chapel Hill, NC, USA,
 * April 2004.
 * 
 * The user can initialize the following values:
 * 1. LP order (lporder)
 * 2. State Transition Model (trn_d)
 * 3. Observation Model (obs_d)
 * 4. process noise covariance (covar_nse_pro_d)
 * 5. observation noise covariance (covar_nse_obs_d)
 * 6. estimate error covariance (covar_nse_err_d)
 * 
 * if order = 1;
 *              item 2 and 3 will be user defined (entered)
 * otherwise;
 *              it will be computed using LP (AR) order.
 *
 * initial estimate is first 'order' observations
 *
 * The algorithm goes as follows:
 *	
 * state_mem_d = initial estimate <- Matrix 
 * filt_data_d Vector storing the final filtered data
 * keep in mind that
 * trn_d, obs_d, covar_nse_pro_d, covar_nse_obs_d, covar_nse_err_d <- Matrix
 *
 * 
 * 1. state_mem_d = trn_d * state_mem_d;
 *
 * 2. covar_nse_err_d = trn_d * covar_nse_err_d * trn_d' + covar_nse_pro_d;
 *
 * 3. k_gain = covar_nse_err_d * obs_d' ;
 *    k_gain = k_gain / 
 *	          (obs_d * covar_nse_err_d * obs_d' + covar_nse_obs_d)^(-1);
 *
 * 4. state_mem_d = state_mem_d 
 *                + k_gain (obsnSequence[index] - obs_d * state_mem_d);
 *
 * 5. covar_nse_err_d = (I - k_gain * obs_d) * covar_nse_err_d;
 *
 * repeat steps til all the data is exhausted,
 * displays the predicted state
 *
 */

//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

/**
 * Class that handles the Kalman filter algorithm
 */
public class AlgorithmKF extends Algorithm {

    //-----------------------------------------------------------------
    //
    // static data members, for interpolation
    //
    //-----------------------------------------------------------------
    int lporder;
    int iporder;
    int scale;

    //-----------------------------------------------------------------
    //
    // primitive data members
    //
    //-----------------------------------------------------------------
    int output_canvas_d[][];   

    //-----------------------------------------------------------------
    //
    // instance data members
    //
    //-----------------------------------------------------------------
    String algo_id = "AlgorithmKF"; 

    /**
     * Original data
     */
    Vector<MyPoint> set_1_d = new Vector<MyPoint>(40, 20);

    /**
     * Stores state estimate results from KF algorithm
     */
    Vector<MyPoint> state_estimate_1_d = new Vector<MyPoint>();
 
    // Interpolation tools -------
    //
    /**
     * for interpolation function, holds original points
     */
    Vector<MyPoint> iset_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * Vector for mean-subtracted [zero-mean] data meaning the original data
     * is interpolated and the average y-value of the set is zero
     */
    Vector<MyPoint> mset_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * For display purpose meaning it contains original data with points
     * in between
     */
    Vector<MyPoint> interpolated_set_1_d = new Vector<MyPoint>(40, 20); 
 
    /**
     * Interpolation tool, for algo PF and KF
     */
    Vector<MyPoint> yestimate_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * to store the average values for each class
     */
    double avg_1_d;

    /**
     * to store autoCorrelation coefficient
     */
    double autoc_1_d[];

    /**
     * to store LP coefficient
     */
    double finallpc_1_d[]; 

    /**
     * to store residual energy 
     */
    double est_err_1_d = 0;	

    /**
     * to store actual error energy
     */
    double actual_err_1_d = 0;

    /**
     * to store the reflection coefficients
     */
    double ref_coeff_1_d[];

    // end interpolation tools ------
    //

    /**
     * kalman gain for each set
     */
    Matrix k_gain_d = new Matrix();
 
     /**
     * observation transition matrix
     */
    Matrix obs_d = new Matrix();

    /**
     * state transition matrix 
     */
    Matrix trn_d = new Matrix();

    /**
     * Process noise variance, Q
     */
    Matrix covar_nse_pro_d = new Matrix();

    /**
     * Measurement noise variance, R
     */
    double covar_nse_obs_d;

    /**
     * Estimate Error variance, P
     */
    Matrix covar_nse_err_d = new Matrix();

    /** 
     * initial state estimate
     */
    Matrix state_mem_d = new Matrix();

    /** 
     * observation sequence 
     * will store the interpolated signal values 
     */
    double[] obs_seq_d;
 
    /**
     * filtered data
     */
    double[] filt_data_d;

    /**
     * index for the observation
     */
    int idx;

    //---------------------------------------------------------------
    //
    // class methods
    //
    //---------------------------------------------------------------

    /**
     * 
     * Implements the initialize() method in the base class. Initializes
     * member data and prepares for execution of first step. This method
     * "resets" the algorithm.
     *
     * @return   returns true if sets of data are valid
     *
     */
    public boolean initialize() {

    point_means_d           = new Vector<MyPoint>();
	description_d           = new Vector<String>();

	yestimate_1_d = new Vector<MyPoint>();
	state_estimate_1_d = new Vector<MyPoint>();

	step_count = 5;
	step_index_d = 0;

	// accepting the LP order and interpolation order
	//
	lporder    = Classify.main_menu_d.lporder;
	iporder    = Classify.main_menu_d.iporder;
	scale      = lporder * iporder;

	// configuring the transition matrix and observation matrix
	//
	trn_d.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	if (lporder == 1) {
	    trn_d.setValue(0, 0, Classify.main_menu_d.trn);
	}
	else {
	    for (int i = 0; i < lporder - 1; i++) {
		int j = i + 1;
		trn_d.setValue(i, j, 1.0);
	    }
	    trn_d.setRow(lporder - 1, 1);
	}

	obs_d.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
	if (lporder == 1) {
	    obs_d.setValue(0, lporder - 1, Classify.main_menu_d.lporder);
	}
	else {
	    obs_d.setValue(0, lporder - 1, 1);
	}
	
	// configure / set noise parameters
	// 
	covar_nse_pro_d.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	covar_nse_err_d.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	covar_nse_pro_d.setDiagonal(Classify.main_menu_d.covar_nse_pro);
	covar_nse_err_d.setDiagonal(Classify.main_menu_d.covar_nse_err);
	covar_nse_obs_d = Classify.main_menu_d.covar_nse_obs;

	// kalman gain parameter dimensions set
	//
	k_gain_d.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);

	// index for the observation sequence initialized
	idx = 0;
	
	// Add the process description for the KF algorithm
	//
	if (description_d.size() == 0) {

	    String str = new String("  0.  Checking for the Data Validity");
	    description_d.addElement(str);
	    
	    str = new String("  1.  Displaying the Input Data Points");
	    description_d.addElement(str);
	    
	    str = new String("  2.  The recursive nature of kalman \n" +
			     "      filtering predicts states point by \n" +
			     "      point as shown in next step. \n" +
			     "\n" +
			     "  State Equation:\n" +
			     "    X(k) = (F) * X(k-1) + v(k) \n"+
			     "  Observation Equation:\n" +
			     "    Y(k) = (M) * X(k) + n(k) \n" );
	    description_d.addElement(str);
	    
	    str = new String("  3.  First iteration");
	    description_d.addElement(str);

	    str = new String("  4.  Second iteration");
	    description_d.addElement(str);

	    str = new String("  5.  Remaining iterations \n");
	    description_d.addElement(str);
	}

	set_1_d = data_points_d.dset1;

	if (checkdata_KF(set_1_d) == true) {
	    pro_box_d.appendMessage((String)description_d.get(step_index_d));
	    pro_box_d.appendMessage("         Data points valid");
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    return true;	
	}
	else {
	    pro_box_d.appendMessage("\n" + "Invalid data");
	    pro_box_d.appendMessage("Clear and enter valid data" + "\n");
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    return false;
	}
    }

    /**
    * 
    * Validates the class entered by user for Kalman Filtering  
    *
    * @param    kf
    * @return   true if data is Vector kf is valid. It is invalid
    *           if the size of kf is 1 or points are not entered
    *           in time progression
    * 
    */
    public boolean checkdata_KF(Vector<MyPoint> kf) {

	if ( kf.size() <= 1) {
	    return false;
	}
	else {

	    for(int i = 0; i <= kf.size() - 1 ; i++ ) {
		for(int j = i + 1; j <= kf.size() - 1 ; j++ ) {

		   // amplitudes should be in time progression
		   //
		   if ( (((MyPoint)kf.elementAt(i)).x == 
			 ((MyPoint)kf.elementAt(j)).x) || 
			(((MyPoint)kf.elementAt(i)).x > 
			 ((MyPoint)kf.elementAt(j)).x) ) {
		       return false;
		   }
		}
	    }
	return true;
	}
    }

    /**
     *
     * Implementation of the run function from the Runnable interface.
     * Determines what the current step is and calls the appropriate method.
     *
     */
    public void run() {

	if (step_index_d == 1) {
	    disableControl();
	    step1();
	    enableControl();
 	}
	if (step_index_d == 2) {
	    disableControl();
	    step2();
	    enableControl(); 
	}
	if (step_index_d == 3) {
	    disableControl();
	    step3(); 
	    enableControl(); 
	}
	if (step_index_d == 4) {
	    disableControl();
	    step4(); 
	    enableControl(); 
	}
	if (step_index_d == 5) {
	    disableControl();
	    step5(); 
	    enableControl(); 
	}

	// exit gracefully
	//
	return;
    }
    
    /**
     *
     * Displays data sets from input box in output box, clears the data
     * points from the vector for input data, zero-mean data,
     * does interpolation, and then finds mean of the interpolated data
     * interpolates the zero-mean data, displays the result
     *
     * @return   True
     *
     */
    boolean step1() {

	// set up progress bar
	//
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);

    	// Display original data
	// 
	if(set_1_d.size() > 0 ) {
       	   interpolated_set_1_d.removeAllElements();
	   iset_1_d.removeAllElements();
	   mset_1_d.removeAllElements();
	   yestimate_1_d.removeAllElements();
	   state_estimate_1_d.removeAllElements();
	   interpol(set_1_d, interpolated_set_1_d);
	   avg_1_d = mean(interpolated_set_1_d, mset_1_d);
	   interpol(mset_1_d, iset_1_d);
	   output_panel_d.addOutput(set_1_d, (Classify.PTYPE_INPUT), 
				    data_points_d.color_dset1);
	}

	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	return true;
    }

   /**
    * find the LP coefficients using levinson durbin algorithm
    *
    * @return    true
    */
    boolean step2() {
	autoCorrelation();
	lpcCoefficient();

	double covar_nse_pro = covar_nse_pro_d.getValue(0, 0);
	double covar_nse_err = covar_nse_err_d.getValue(0, 0);
	pro_box_d.appendMessage(new String(
			     " Class 0: \n\n" +
			     "    Variances \n" +
			     "     Q = process noise      = " + 
			     MathUtil.SetDecimal(covar_nse_pro, 4) + "\n" +
			     "     R = observation noise  = " + 
			     MathUtil.SetDecimal(covar_nse_obs_d, 4) + "\n" +
			     "     P = estimate error     = " + 
			     MathUtil.SetDecimal(covar_nse_err, 4) + "\n"));
	return true;
    }   

   /**
    * Used to display and compute the first iteration of set_1_d
    *
    * @return    true
    *
    */
    boolean step3() {

	double est_obsn = 0;
 	String message = new String();

	int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(4);

	scale = iset_1_d.size();
	obs_seq_d = new double[scale];
	filt_data_d = new double[scale];

	if (set_1_d.size() > 0) {
	    for (int i = 0; i < scale; i++) {
		obs_seq_d[i] = iset_1_d.elementAt(i).y;
	    }

	    state_mem_d.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	    for ( ; idx < lporder; idx++) {
		double data_pt = 
		    ((MyPoint)iset_1_d.elementAt(idx)).y;

		state_mem_d.setValue(idx, 0, data_pt);

		MyPoint p = new MyPoint();
		p.y = ((MyPoint)iset_1_d.elementAt(idx)).y + avg_1_d;
		p.x = ((MyPoint)iset_1_d.elementAt(idx)).x;
		yestimate_1_d.addElement(new MyPoint(p));
	    }
	}

	for (int i = 0; i < lporder; i++) {
	    double value = -finallpc_1_d[lporder - i];
	    trn_d.setValue(lporder - 1, i, value);
	}

	// for (int i = 0; i < scale - lporder; i++) {
	for (int i = 0; i < 2; i++) {
	 
	    if(set_1_d.size() > 0) {

		// Equation 1 -> x(k) = F * x(k-1)
		//
		Matrix mat_dummy = new Matrix();
		mat_dummy.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		trn_d.multMatrix(state_mem_d, mat_dummy);
		state_mem_d.copyMatrix(mat_dummy);

		// Equation 2 -> P(k) = F * P(k-1) * F' + Q
		// mat1: for transpose
		// mat2: for adhoc -temporary reasons
		//
		Matrix mat1 = new Matrix();
		mat1.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

		Matrix mat2 = new Matrix();
		mat2.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

		trn_d.transposeMatrix(mat1);
		trn_d.multMatrix(covar_nse_err_d, mat2);

		mat_dummy.copyMatrix(mat2);
		mat2.multMatrix(mat1, mat_dummy);
		mat2.copyMatrix(mat_dummy);
		mat2.addMatrix(covar_nse_pro_d);
		covar_nse_err_d.copyMatrix(mat2);

		// Equation 3 -> k(k) = P(k) * H' * (H * P(k) * H' + R)^(-1)
		//
		mat1.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		obs_d.transposeMatrix(mat1);

		mat_dummy.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		covar_nse_err_d.multMatrix(mat1, mat_dummy);
		mat1.copyMatrix(mat_dummy);

		Matrix mat3 = new Matrix();
		mat3.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
		mat3.copyMatrix(obs_d);
		mat_dummy.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
		mat3.multMatrix(covar_nse_err_d, mat_dummy);
		mat3.copyMatrix(mat_dummy);

		Matrix mat4 = new Matrix();
		mat4.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		obs_d.transposeMatrix(mat4);

		Matrix mat5 = new Matrix();
		mat5.initMatrixValue(1, 1, 0.0, Matrix.FULL);

		mat3.multMatrix(mat4, mat5);

		double value = mat5.getValue(0, 0);
		value = value + covar_nse_obs_d;
		value = 1.0 / value;
		mat1.scalarMultMatrix(value);

		k_gain_d.copyMatrix(mat1); 

		// Equation 4 -> x(k) = x(k) + k(k) * ( obsn(i) - H * x(k))
		//
		est_obsn = state_mem_d.getValue(lporder - 1, 0);
		est_obsn = obs_seq_d[idx] - est_obsn;
		k_gain_d.scalarMultMatrix(est_obsn);
		state_mem_d.addMatrix(k_gain_d);

		// Equation 5 -> P(k) = ( I - k(k) * H) * P(k)
		//
		mat1.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		mat1.identityMatrix();

		mat2.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		k_gain_d.multMatrix(obs_d, mat2);

		mat_dummy.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		mat1.subtractMatrix(mat2, mat_dummy);
		mat1.copyMatrix(mat_dummy);

		mat_dummy.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		covar_nse_err_d.multMatrix(mat1, mat_dummy);
		covar_nse_err_d.copyMatrix(mat_dummy);

		// filtered data
		// 

		MyPoint p = new MyPoint();
		p.y = ((MyPoint)interpolated_set_1_d.elementAt(idx)).y;
		p.x = ((MyPoint)interpolated_set_1_d.elementAt(idx)).x;
		yestimate_1_d.addElement(new MyPoint(p));

		yestimate_1_d.elementAt(idx).y = 
		    state_mem_d.getValue(lporder - 1, 0) + avg_1_d ;
		idx++;

		
		output_panel_d.addOutput(yestimate_1_d,
					 Classify.PTYPE_LINE,
					 Color.BLACK);
		output_panel_d.addOutput(yestimate_1_d,
					 (Classify.PTYPE_INPUT) * 4,
					 Color.BLACK);
		output_panel_d.repaint();
		
		try { Thread.sleep(50);
		}
		catch (Exception e) {}
		
		// removes the last output 
		//
		output_panel_d.removeOutput();
		output_panel_d.removeOutput();
		pro_box_d.setProgressCurr(++progress_curr);
	    }

	} // end first iteration, set 1

	if (set_1_d.size() == 0) {
	    message = new String("    Class 0 - data not entered \n");
	    pro_box_d.appendMessage(message);
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    pro_box_d.setProgressCurr(++progress_curr);
	}

	pro_box_d.setProgressCurr(4);
	return true;
    }   

   /**
    * Used to display and compute the second iteration of set_1_d
    *
    * @return    true
    *
    */
    boolean step4() {
	double est_obsn = 0;
	String message = new String();
	int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(4);

	if(set_1_d.size() > 0) {
		// Equation 1 -> x(k) = F * x(k-1)
		//
		Matrix mat_dummy = new Matrix();
		mat_dummy.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		trn_d.multMatrix(state_mem_d, mat_dummy);
		state_mem_d.copyMatrix(mat_dummy);

		// Equation 2 -> P(k) = F * P(k-1) * F' + Q
		// mat1: for transpose
		// mat2: for adhoc -temporary reasons
		//
		Matrix mat1 = new Matrix();
		mat1.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

		Matrix mat2 = new Matrix();
		mat2.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

		trn_d.transposeMatrix(mat1);
		trn_d.multMatrix(covar_nse_err_d, mat2);

		mat_dummy.copyMatrix(mat2);
		mat2.multMatrix(mat1, mat_dummy);
		mat2.copyMatrix(mat_dummy);
		mat2.addMatrix(covar_nse_pro_d);
		covar_nse_err_d.copyMatrix(mat2);

		// Equation 3 -> k(k) = P(k) * H' * (H * P(k) * H' + R)^(-1)
		//
		mat1.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		obs_d.transposeMatrix(mat1);

		mat_dummy.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		covar_nse_err_d.multMatrix(mat1, mat_dummy);
		mat1.copyMatrix(mat_dummy);

		Matrix mat3 = new Matrix();
		mat3.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
		mat3.copyMatrix(obs_d);
		mat_dummy.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
		mat3.multMatrix(covar_nse_err_d, mat_dummy);
		mat3.copyMatrix(mat_dummy);

		Matrix mat4 = new Matrix();
		mat4.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		obs_d.transposeMatrix(mat4);

		Matrix mat5 = new Matrix();
		mat5.initMatrixValue(1, 1, 0.0, Matrix.FULL);

		mat3.multMatrix(mat4, mat5);

		double value = mat5.getValue(0, 0);
		value = value + covar_nse_obs_d;
		value = 1.0 / value;
		mat1.scalarMultMatrix(value);

		k_gain_d.copyMatrix(mat1); 

		// Equation 4 -> x(k) = x(k) + k(k) * ( obsn(i) - H * x(k))
		//
		est_obsn = state_mem_d.getValue(lporder - 1, 0);
		est_obsn = obs_seq_d[idx] - est_obsn;
		k_gain_d.scalarMultMatrix(est_obsn);
		state_mem_d.addMatrix(k_gain_d);

		// Equation 5 -> P(k) = ( I - k(k) * H) * P(k)
		//
		mat1.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		mat1.identityMatrix();

		mat2.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		k_gain_d.multMatrix(obs_d, mat2);

		mat_dummy.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		mat1.subtractMatrix(mat2, mat_dummy);
		mat1.copyMatrix(mat_dummy);

		mat_dummy.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		covar_nse_err_d.multMatrix(mat1, mat_dummy);
		covar_nse_err_d.copyMatrix(mat_dummy);

		// filtered data
		// 

		MyPoint p = new MyPoint();
		p.y = ((MyPoint)interpolated_set_1_d.elementAt(idx)).y;
		p.x = ((MyPoint)interpolated_set_1_d.elementAt(idx)).x;
		yestimate_1_d.addElement(new MyPoint(p));

		yestimate_1_d.elementAt(idx).y = 
		    state_mem_d.getValue(lporder - 1, 0) + avg_1_d ;
		idx++;

	    output_panel_d.addOutput(yestimate_1_d, Classify.PTYPE_LINE,
					 Color.BLACK);
	    output_panel_d.addOutput(yestimate_1_d,
				     (Classify.PTYPE_INPUT) * 4,
				     Color.BLACK);
	    output_panel_d.repaint();
	    
	    try {
		
		Thread.sleep(50);
	    }
	    catch (Exception e) {}
	    
	    // Removes the last output 
	    //
	    output_panel_d.removeOutput();
	    pro_box_d.setProgressCurr(++progress_curr);
	    
	} // end second iteration, set 1
	else {
	    message = new String("    Class 0 not entered \n");
	    pro_box_d.appendMessage(message);
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    pro_box_d.setProgressCurr(++progress_curr);
	}
	pro_box_d.setProgressCurr(4);
	return true;
    }   
 
    /**
     *
     *	
     * 1. newState = trn_d * currentState;
     *
     * 2. varErrorNew = varErrorInitial * trn_d * trn_d + varStateNoise;
     *    estimateObsn = obs_d * newState;
     *
     * 3. k_gain = obs_d * varErrorNew ;
     *	  k_gain = k_gain / 
     *	          ( obs_d * obs_d * varErrorNew + varObsnNoise);
     *
     * 4. newerState =
     *	          newState + k_gain * (obsnSequence[index] - estimateObsn);
     *
     * 5. double a = 0;
     * 	  a = ( 1 - k_gain * obs_d) * ( 1 - k_gain * obs_d);
     *
     *	varErrorNewer =
     *             varErrorNew * a + varObsnNoise * k_gain * k_gain;
     *
     *	varErrorInitial = varErrorNewer;
     *	currentState = newerState; 
     *
     * displays the predicted signal
     *
     * @return   true
     * 
     */
    boolean step5() {
	double est_obsn = 0;
	// int total_pts = interpolated_set_1_d.size();
	int progress_curr = 0;

	pro_box_d.setProgressMax(scale - idx);
	pro_box_d.setProgressCurr(0);

	//for (int i = idx; i < scale - lporder; i++) {
	for (int i = idx; i < scale; i++) {
	    if(set_1_d.size() > 0) {

		// Equation 1 -> x(k) = F * x(k-1)
		//
		Matrix mat_dummy = new Matrix();
		mat_dummy.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		trn_d.multMatrix(state_mem_d, mat_dummy);
		state_mem_d.copyMatrix(mat_dummy);

		// Equation 2 -> P(k) = F * P(k-1) * F' + Q
		// mat1: for transpose
		// mat2: for adhoc -temporary reasons
		//
		Matrix mat1 = new Matrix();
		mat1.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

		Matrix mat2 = new Matrix();
		mat2.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

		trn_d.transposeMatrix(mat1);
		trn_d.multMatrix(covar_nse_err_d, mat2);

		mat_dummy.copyMatrix(mat2);
		mat2.multMatrix(mat1, mat_dummy);
		mat2.copyMatrix(mat_dummy);
		mat2.addMatrix(covar_nse_pro_d);
		covar_nse_err_d.copyMatrix(mat2);

		// Equation 3 -> k(k) = P(k) * H' * (H * P(k) * H' + R)^(-1)
		//
		mat1.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		obs_d.transposeMatrix(mat1);

		mat_dummy.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		covar_nse_err_d.multMatrix(mat1, mat_dummy);
		mat1.copyMatrix(mat_dummy);

		Matrix mat3 = new Matrix();
		mat3.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
		mat3.copyMatrix(obs_d);
		mat_dummy.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
		mat3.multMatrix(covar_nse_err_d, mat_dummy);
		mat3.copyMatrix(mat_dummy);

		Matrix mat4 = new Matrix();
		mat4.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		obs_d.transposeMatrix(mat4);

		Matrix mat5 = new Matrix();
		mat5.initMatrixValue(1, 1, 0.0, Matrix.FULL);

		mat3.multMatrix(mat4, mat5);

		double value = mat5.getValue(0, 0);
		value = value + covar_nse_obs_d;
		value = 1.0 / value;
		mat1.scalarMultMatrix(value);

		k_gain_d.copyMatrix(mat1); 

		// Equation 4 -> x(k) = x(k) + k(k) * ( obsn(i) - H * x(k))
		//
		est_obsn = state_mem_d.getValue(lporder - 1, 0);
		est_obsn = obs_seq_d[idx] - est_obsn;
		k_gain_d.scalarMultMatrix(est_obsn);
		state_mem_d.addMatrix(k_gain_d);

		// Equation 5 -> P(k) = ( I - k(k) * H) * P(k)
		//
		mat1.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		mat1.identityMatrix();

		mat2.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		k_gain_d.multMatrix(obs_d, mat2);

		mat_dummy.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		mat1.subtractMatrix(mat2, mat_dummy);
		mat1.copyMatrix(mat_dummy);

		mat_dummy.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
		covar_nse_err_d.multMatrix(mat1, mat_dummy);
		covar_nse_err_d.copyMatrix(mat_dummy);

		// filtered data
		// 

		MyPoint p = new MyPoint();
		p.y = ((MyPoint)interpolated_set_1_d.elementAt(idx)).y;
		p.x = ((MyPoint)interpolated_set_1_d.elementAt(idx)).x;
		yestimate_1_d.addElement(new MyPoint(p));

		yestimate_1_d.elementAt(idx).y = 
		    state_mem_d.getValue(lporder - 1, 0) + avg_1_d ;
		idx++;

		output_panel_d.addOutput(yestimate_1_d,
					 Classify.PTYPE_LINE,
					 Color.black);
		output_panel_d.addOutput(yestimate_1_d,
					 (Classify.PTYPE_INPUT) * 4,
					 Color.black);
		output_panel_d.repaint();

		try { Thread.sleep(50);
		}
		catch (Exception e) {}
		
		// Removes the last output 
		//
		output_panel_d.removeOutput();
		pro_box_d.setProgressCurr(++progress_curr);
	    }
	}

	if (set_1_d.size() > 0) {
	    
	    double error = actual_error(yestimate_1_d, 
					interpolated_set_1_d);
	    pro_box_d.appendMessage(" Class 0: \n" +
				    "  Mean Square Error   = " +
				    MathUtil.SetDecimal(error, 6) + "\n");
	}
	pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	output_panel_d.repaint();
	pro_box_d.appendMessages("Algorithm Complete " + "\n");

       	return true;
    }
 
    /**
     *
     * Calculates the interpolated points for the data inputs   
     *
     * @param  v  input data points
     * @param  iset  interpolated data points
     *
     */
    public void interpol(Vector<MyPoint> v, Vector<MyPoint> iset)
    {	
	double sample;
	double d ;
	MyPoint r = new MyPoint();

	double[] y2 = new double[v.size()];  
	double[] y  = new double[v.size()];
	double[] x  = new double[v.size()];
       
       for (int k = 0; k < v.size(); k++)
       {
	   x[k] = ((MyPoint)v.elementAt(k)).x;
	   y[k] = ((MyPoint)v.elementAt(k)).y;
       }

       // Reference:
       // "Numerical Recipe in C", Cambridge University Press, pp 113-116
       //
       // total time interval of the data points calculated
       //
       d = ((MyPoint)v.lastElement()).x - ((MyPoint)v.firstElement()).x ; 
       
       // time step calculation
       //
       sample = d / scale;
       spline(x, y, y2, v.size());
       
       // iset holds new data points - interpolated values
       // First value added 
       //
       iset.addElement(new MyPoint((MyPoint)v.firstElement()));
       
       // Increment to next Time value
       //
      	r.x = sample + ((MyPoint)v.firstElement()).x;

	while (r.x <=  ((MyPoint)v.lastElement()).x)
	{
	    for(int j = 1; j < v.size(); j++)
	    {
		// the New Interpolated amplitude will be calculated
		// for the two points within which the time intervals 
		// lies, Hence, the condition check
		//
		if((r.x <= ((MyPoint)v.elementAt(j)).x ) && 
		   (r.x >=  ((MyPoint)v.elementAt(j - 1)).x)) 
		{
		    r.y = 0 ;
		    splint ((MyPoint)v.elementAt(j - 1), 
			    (MyPoint)v.elementAt(j), r, y2, (j - 1));

		    // New value appended to the array
		    //
		    iset.addElement(new MyPoint(r));
		    break; 
		}
	    }	    
   	    r.x = r.x + sample;
	}
	// sometimes the last element of the user entered 
	// values is missed out hence the condition to 
	// cross-check and put it in
	//
	if (((MyPoint)iset.lastElement()).x != ((MyPoint)v.lastElement()).x )
	    iset.addElement( new MyPoint((MyPoint)v.lastElement()));
    }

    /**
     *
     * Actually interpolates the points
     *
     * @param x    array containing the x coordinates of datapoints
     * @param y    array containing the y coordinates of datapoints
     * @param y2   array containing the interpolated y coordinates
     * @param size the size of the array to be interpolated
     * 
     */
    public void spline (double[] x, double[] y, double[] y2, int size )
    { 
	double[] u = new double[size];
	double p; 
	double sig; 
	
	y2[0] = 0;
	u[0] = 0;
	 
	for( int i = 1; i < size - 1; i++ )
	{ 
	    sig = ( x[i] - x[i - 1] ) / ( x[i + 1] - x[i - 1] ) ;
	    p = sig * y2[i - 1] + 2 ;
	    y2[i] = ( sig - 1) / p ;
	    u[i] = ( y[i + 1] - y[i] ) / ( x[i + 1] - x[i] ) ;
	    u[i] = ( 6 * u[i] / ( x[i + 1] - x[i - 1] ) 
		     - sig * u[i - 1] ) / p ;
	}
	y2[size - 1] = 0 ;
	
	// This is the back substitution loop of the tridiagonal algorithm
	//
	for(int i = size - 2; i >= 0; i-- )     
	    y2[i] = y2[i] * y2[i + 1] + u[i] ;
    }

    /**
     * Interpolates for a point between the two known points
     * using Cubic Interpolation
     *
     * @param   u1 start point for the interpolation
     * @param   u2 end point for the interpolation
     * @param   r  returning point, basically the interpolated point
     * @param   y2 array used for reassigning of r
     * @param   i  the sample number
     *
     */
    public void splint ( MyPoint u1, MyPoint u2, MyPoint r, 
			 double[] y2, int i )
    {
	double h;
	double a;
	double b;
 
	h = u2.x - u1.x ;
	a = ( u2.x - r.x ) / h ;
	b = ( r.x - u1.x ) / h ;
	r.y = a * u1.y + b * u2.y + ((a * a * a - a) * y2[i] 
	                             + ( b * b * b - b ) * y2[i + 1] ) 
	                                                 * ( h * h ) / 6 ; 
    }

    /**
     *
     * Calculates the mean and the zero-mean data points  
     *
     * @param    v orginal datapoints
     * @param    mv zero mean datapoints 
     *
     * @return   The average in a double
     * 
     */
    public double mean(Vector<MyPoint> v, Vector<MyPoint> mv)
    {
        double average = 0;
	MyPoint p      = new MyPoint();

       for(int i = 0; i < v.size(); i++ )
	   average = average + ((MyPoint)v.elementAt(i)).y;
       average = average / v.size();
       
       for(int i = 0; i < v.size(); i++ )
       {
	   p.y = ((MyPoint)v.elementAt(i)).y - average;
	   p.x =  ((MyPoint)v.elementAt(i)).x; 
	   mv.addElement(new MyPoint(p));
       }        
       return average ;
    }  
   
    /**
    *
    * Computes the autocorrelation coeffient from the data sets   
    *
    */
    public void autoCorrelation()
    {
	if (iset_1_d.size() > 0 )
	{
	    autoc_1_d = new double[lporder + 1];
	    autocorrelate(iset_1_d, autoc_1_d);
	}	
    }

    /**
     *
     * Actaully computes the autocorrelation coefficients
     *
     * @param  v Vector of datapoints
     * @param  autoCoeff_co array of autocorrelation coefficients
     *
     */
    public void autocorrelate ( Vector v, double[] autoCoeff_co) 
    {
	int size;
	size                 = v.size();
      	double[] amplitude_y = new double[size];

      	// to store the amplitude for calculation autocoefficient
      	//
    	for (int i = 0; i < size; i++)
	    amplitude_y[i] = ((MyPoint)v.elementAt(i)).y;

    	for (int j = 0; j <= lporder; j++)
      	{
      	    autoCoeff_co[j] = 0;
      	    if ( j < size)
      		for (int k = 0; k < size - j; k++)
		    autoCoeff_co[j] = autoCoeff_co[j] 
			            + amplitude_y[k] * amplitude_y[k + j];
	}

   	// normalize the autocorrelation cofficient
      	//
     	for (int i = lporder; i >= 0; i--) 
      	    autoCoeff_co[i] = autoCoeff_co[i] / autoCoeff_co[0];
    }

    /**
     *
     * Computes the Linear Prediction coefficient from the data sets   
     *
     */
    public void lpcCoefficient()
    {
	// Rabiner, Schafer, "Digital Processing of Speech signals", 
	// Bell Laboratories, Inc. 1978, 
	// section 8.3.2 Durbin's recursive solution for 
	// the autocorrelation equations pp 411-412
	//
	if (set_1_d.size() > 0)
	{
	    finallpc_1_d = new double [lporder + 1];
	    ref_coeff_1_d = new double [lporder + 1];
	    est_err_1_d = calculate_lpc(autoc_1_d, 
					    finallpc_1_d, ref_coeff_1_d);
	}
    }

    /**
     *
     * Actually calculate the LP coefficient and the Residual Error 
     * Energy, and Reflection Coefficients
     *
     * @param  auto_coeff array of auto correlation coefficients
     * @param  lpc        array of linear prediction coefficients
     * @param  rc_reg     array of reflection coefficients
     * 
     * @return residual error energy in a double
     *
     */
    public double calculate_lpc (double[] auto_coeff, 
				 double[] lpc, double[] rc_reg)
    {
	int j_bckwd       = 0;
	long middle_index = 0;

	// Reference:
	// J.D. Markel and A.H. Gray, "Linear Prediction of Speech,"
	// Springer-Verlag Berlin Heidelberg, New York, USA, pp. 219,
	// 1980.
 
	// initialization
	//
	lpc[0] = 1;
	double err_energy = auto_coeff[0];

	// if the energy of the signal is zero we set all the predictor
	// coefficients to zero and exit
	//
	if (err_energy == (double)0)
	{
	    for (int i = 0; i < lporder + 1 ; i++)
		lpc[i] = 0;
	}
	else
        {
	    // do the first step manually
	    //
	    double sum = 0;
	    double tmp = 0;
	    
	    sum         = -auto_coeff[1] / err_energy;
	    rc_reg[1]   = sum;
	    lpc[1]      = rc_reg[1];
	    tmp         = 1 - rc_reg[1] * rc_reg[1];
	    err_energy *= tmp;
	    // recursion
	    //
	    for (int i = 2; i <= lporder; i++)
	    {
		sum = 0;
		for (int j = 1; j < i; j++)
		    sum += lpc[j] * auto_coeff[i - j];
 
		rc_reg[i]    = -(auto_coeff[i] + sum) / err_energy;	    
		lpc[i]       = rc_reg[i];
		j_bckwd      = i - 1;
		middle_index = i / 2;
		for (int j = 1; j <= middle_index; j++)
		{
		    sum          = lpc[j_bckwd] + rc_reg[i] * lpc[j];
		    lpc[j]       = lpc[j] + rc_reg[i] * lpc[i - j];
		    lpc[j_bckwd] = sum;
		    j_bckwd--;
		}
		// compute new error
		//
		tmp = 1.0 - rc_reg[i] * rc_reg[i];
		err_energy *= tmp;
	    }
	}
	return err_energy;
    }

    /**
     *
     * Compute the actual error from the given data points and the estimated
     * values.
     *
     * @param  y_estimate datapoints of the estimated datapoints
     * @param  iset       original datapoints 
     *
     * @return actual error energy in a double
     */
    public double actual_error (Vector y_estimate, Vector iset)
    {
	double error_value;
	double act_error = (double) 0;
	int j = 0;
	int i = 0;

	// first point and the last point have their x-coordinates same
	// So the difference in y-values is the error value
	//
	error_value = ((MyPoint)y_estimate.firstElement()).y 
	              - ((MyPoint)iset.firstElement()).y;
	act_error = act_error + error_value * error_value; 
	error_value = ((MyPoint)y_estimate.lastElement()).y 
	              - ((MyPoint)iset.lastElement()).y;
	act_error = act_error + error_value * error_value;

	// for next values
	// need to continue till all the user defined points are exhausted
	//
	j++;

	for (i = 1; ( (i < y_estimate.size()) && (j < iset.size()) ); i++)
	{ 
	    while ( (((MyPoint)y_estimate.elementAt(i)).x < 
		     ((MyPoint)iset.elementAt(j)).x) 
		    && ((i < y_estimate.size() - 1) && (j < iset.size())) )
	    {
		    i++;
	    }

	    if ( j < iset.size())
	    {
		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     == ((MyPoint)iset.elementAt(j)).x ) 
		{
		    error_value = ((MyPoint)y_estimate.elementAt(i)).y 
			          - ((MyPoint)iset.elementAt(j)).y;
		    act_error = act_error + error_value * error_value;
		}

		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     > ((MyPoint)iset.elementAt(j)).x )
		{
		    double y1 = ((MyPoint)y_estimate.elementAt(i)).y;
		    double y2 = ((MyPoint)y_estimate.elementAt(i - 1)).y;
		    double x1 = ((MyPoint)y_estimate.elementAt(i)).x;
		    double x2 = ((MyPoint)y_estimate.elementAt(i - 1)).x;
		    double x_unknown = ((MyPoint)iset.elementAt(j)).x;
		    error_value = y2 - ( (x2 - x_unknown) * 
					 ( y2 - y1) / (x2 - x1));
		    act_error = act_error + error_value * error_value ;
		}
	    }
	    j++;
	}
	return act_error;
    }
    
} //end class
