/*
 * AlgorithmPF.java v6.1 
 * Created by: Sanjay Patil 01/10/2006
 * modified: Sanjay Patil 01/17/2006
 * 
 * Description: Particle Filtering
 *
 *
 * Three step algorithm:
 *
 * 1. Compute predict states
 * 2. Importance weights
 * 3. Update states
 *
 * ParticleFilter: a class that computes the states (features) of a
 * signal using particle filter.
 *
 * particle filters are based on sequential Monte Carlo (SMC)
 * technique. it is a technique for implementaing a recursive Bayesian
 * filter by MC simulations.

 *
 * particle filters are used to model signals using sequential prediction
 * techniques. a particle filter is represented by to state space equations:
 *
 * state equation:       x(k) = F(x(k-1), u(k-1))
 * observation equation: y(k) = H(x(k), v(k))
 *
 * where:
 *  k -> time instant
 *  x -> hidden (to find) state vector
 *  y -> observation (measurement) vector
 *  u -> process noise
 *  v -> observation noise
 *  F -> state transition model
 *  H -> observation model
 *
 * process noise and observation noise can be have either Gaussian or
 * nonGaussian statistics.
 * state transition model and observation model can be either linear
 * or nonlinear and also both of these models can be time-varying.
 *
 * particle filtering consists of two steps:
 *
 * step 1:
 * 
 *  p(x(k)|z(1:k-1)) = (integration) p(x(k)|x(k-1))  p(x(k-1)|y(k-1) dxk-1
 *
 * step 2: at time step k, as measurement y(k) becomes available,
 * update stage is executed
 *
 *                          p(y(k)|x(k))  p(x(k)|y(1:k-1))
 *  p(x(k)k|y(1:k)) = ------------------------------------------
 *                                p(y(k)|y(1:k-1))
 *
 *  normalizing constant
 *
 *  p(y(k)|y(1:k-1)) =  (integration) p(y(k)|x(k))  p(x(k)|y(1:k-1)) dxk
 *
 *  where:
 *
 *  p(y(k)|x(k)) is known from observation equation and statistics of v(k)
 *  read p(a|b) as probability of 'a' given 'b'
 *
 * for our implementation, we have considered the following:
 * 1. x    -> a vector
 * 2. y    -> a scalar time series
 * 3. u, v -> Gaussian zero mean noise with statistics as
 *            u -> N(0, Q); v -> N(0, R)
 * 4. F, H -> linear AR models, hence matrices.
 * 
 * such a implementation would be termed as particle kalman filter.
 * also refer KalmanFilter class for more details on kalmanfilter
 * implementation
 *
 * the implementations in this class follow:
 *
 * N. Arulampalam, S. Maskell, N. Gordan, and T. Clapp, "Tutorial On
 * Particle Filters For Online Nonlinear/ Non-Gaussian Bayesian
 * Tracking," IEEE Transactions on Signal Processing, vol. 50, no. 2,
 * pp. 174-188, February 2002.
 *
 *  R. van der Merve, N. de Freitas, A. Doucet, and E. Wan, "The
 *  Unscented Particle Filter," Technical Report CUED/F-INFENG/TR 380,
 *  Cambridge University Engineering Department, Cambridge University,
 *  U.K., August 2000.
 *
 * The user can initialize the following values:
 * 1. LP order (lporder)
 * 2. State Transition Model (trn_d)
 * 3. Observation Model (obs_d)
 * 4. process noise covariance (covar_nse_pro_d)
 * 5. observation noise covariance (covar_nse_obs_d)
 * 6. estimate error covariance (covar_nse_err_d)
 * 
 * if order = 1;
 *              item 2 and 3 will be user defined (entered)
 * otherwise;
 *              it will be computed using LP (AR) order.
 *
 * initial estimate is first 'order' observations
 *
 * The algorithm goes as follows:
 *	
 * state_mem_d = initial estimate <- Matrix 
 * filt_data_d Vector storing the final filtered data
 * keep in mind that
 * trn_d, obs_d, covar_nse_pro_d, covar_nse_obs_d, covar_nse_err_d <- Matrix
 *
 */

//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

/**
 * This class implements the particle filtering algorithm
 */
public class AlgorithmPF extends Algorithm {

    //-----------------------------------------------------------------
    //
    // static data members
    //
    //-----------------------------------------------------------------
    int lporder;
    int iporder;
    int scale;

    //-----------------------------------------------------------------
    //
    // primitive data members
    //
    //-----------------------------------------------------------------
    int output_canvas_d[][];   

    //-----------------------------------------------------------------
    //
    // instance data members
    //
    //-----------------------------------------------------------------

    String algo_id = "AlgorithmPF"; 
  
    /**
     * for original data
     */
    Vector<MyPoint> set_1_d = new Vector<MyPoint>();
 
     /**
     * Stores state estimate results from PF algorithm
     */
    Vector<MyPoint> state_estimate_1_d = new Vector<MyPoint>();

    // Interpolation tools -------
    //
    /**
     * for interpolation function, holds original points
     */
    Vector<MyPoint> iset_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * Vector for mean-subtracted [zero-mean] data meaning the original data
     * is interpolated and the average y-value of the set is zero
     */
    Vector<MyPoint> mset1_d = new Vector<MyPoint>(40, 20); 
 
    /**
     * For display purpose meaning it contains original data with points
     * in between
     */
    Vector<MyPoint> interpolated_set_1_d = new Vector<MyPoint>(40, 20); 

     /**
     * Used for interpolation
     */
    Vector<MyPoint> yestimate_1_d = new Vector<MyPoint>(); 

   /**
     * to store the average values for each class
     */
    double avg_1_d;

    /**
     * to store autoCorrelation coefficient
     */
    double autoc_1_d[];

    /**
     * to store LP coefficient
     */
    double finallpc_1_d[]; 

    /**
     * to store the reflection coefficients
     */
    double ref_coeff_1_d[];

     /**
     * to store residual energy 
     */
    double est_err_1_d = 0;	

    /**
     * to store actual error energy
     */
    double actual_err_1_d = 0;

   // end interpolation tools ------
    //

     /**
     * observation transition matrix
     */
    Matrix obs_d = new Matrix();

    /**
     * state transition matrix 
     */
    Matrix trn_d = new Matrix();

    /**
     * Number of Gaussian samples
     */
    int particles_d;

    /**
     * Measurement noise variance, R
     */
    double covar_nse_obs_d;

    /**
     * Process noise variance, Q
     */
    double covar_nse_pro_d;

    /**
     * Initial variance of state estimate
     */
    double covar_nse_err_d;

    /** 
     * state matrix
     */
    Matrix state_mem_d = new Matrix();

    /**
     * observation sequence
     * will store the interpolated signal values
     */
    double[] obsn_seq_d;

    /** 
     * index for the observation
     */
    int idx;

    //---------------------------------------------------------------
    //
    // public class methods
    //
    //---------------------------------------------------------------

    /**
     * 
     * Implements the initialize() method in the base class. Initializes
     * member data and prepares for execution of first step. This method
     * "resets" the algorithm.
     *
     * @return   returns true if sets of data are valid
     *
     */
    public boolean initialize() {

	
	point_means_d           = new Vector<MyPoint>();
	description_d           = new Vector<String>();

	yestimate_1_d = new Vector<MyPoint>();
	state_estimate_1_d = new Vector<MyPoint>();

	// setting up all the parameters and initializing them
	//
	step_count = 5;
	step_index_d = 0;

	particles_d = Classify.main_menu_d.particles;
	lporder    = Classify.main_menu_d.lporder;
	iporder    = Classify.main_menu_d.iporder;
	scale      = lporder * iporder;

	// initializing transition and observation matrix
	//
	trn_d.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	if (lporder == 1) {
	    trn_d.setValue(0, 0, Classify.main_menu_d.trn);
	}
	else {
	    for (int i = 0; i < lporder - 1; i++) {
		int j = i + 1;
		trn_d.setValue(i, j, 1.0);
	    }
	    trn_d.setRow(lporder - 1, 1);
	}

	obs_d.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
	if (lporder == 1) {
	    obs_d.setValue(0, lporder - 1, Classify.main_menu_d.lporder);
	}
	else {
	    obs_d.setValue(0, lporder - 1, 1);
	}

	// initializing noise parameters
	//
	covar_nse_pro_d = Classify.main_menu_d.covar_nse_pro;
	covar_nse_err_d = Classify.main_menu_d.covar_nse_err;
	covar_nse_obs_d = Classify.main_menu_d.covar_nse_obs;

	// index for the observation sequence initialized
	//
	idx = 0;

	// Add the process description for the PF algorithm
	//
	if (description_d.size() == 0) {

	    String str = new String("  0.  Checking for the Data Validity");
	    description_d.addElement(str);
	    
	    str = new String("  1.  Displaying Input Data Points");
	    description_d.addElement(str);
	    
	    str = new String("  2.  The recursive nature of particle \n" +
			     "      filtering predicts states point by \n" +
			     "      point as shown in next step. \n" +
			     "\n" +
			     "  State Equation:\n" +
			     "    X(k) = (F) * X(k-1) + v(k) \n" +
			     "  Observation Equation:\n" +
			     "    Y(k) = (M) * X(k-1) + n(k) \n" +
			     "\n" +
			     "  F = Transition model  = \n" +
			     "           P( x(k) | x(k-1) )\n" +
			     "  M = Observation model = \n" +
			     "           P( y(k) | y(k-1) )\n" +
			     "  Number of Particles = " + 
  			           particles_d + "\n" );
			     
	    description_d.addElement(str);
	    
	    str = new String("  3.  First iteration");
	    description_d.addElement(str);

	    str = new String("  4.  Second iteration");
	    description_d.addElement(str);

	    str = new String("  5.  Remaining iterations \n");
	    description_d.addElement(str);
 	}

	set_1_d = data_points_d.dset1;

	if(checkdata_PF(set_1_d) == true) {
	    pro_box_d.appendMessage((String)description_d.get(step_index_d));
	    pro_box_d.appendMessage("         Data points valid");
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    return true;	
	}
	else  {
	    pro_box_d.appendMessage("\n" + "Invalid data");
	    pro_box_d.appendMessage("Clear and enter valid data" + "\n");
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    return false;
	}
    }
    
    /**
     * 
     * Validates the class entered by user for Partical Filtering  
     *
     * @param    pf
     * @return   true if data is Vector pf is valid. It is invalid
     *           if the size of pf is 1 or points are not entered in
     *           time progression
     * 
     */
    public boolean checkdata_PF(Vector<MyPoint> pf) {

	if ( pf.size() <= 1) {
	    return false;
	}
	else {
	    for(int i = 0; i <= pf.size() - 1 ; i++ ) {
		for(int j = i + 1; j <= pf.size() - 1 ; j++ ) {

		   // amplitudes should be in time progression
		   //
		   if ( (((MyPoint)pf.elementAt(i)).x == 
			 ((MyPoint)pf.elementAt(j)).x) || 
			(((MyPoint)pf.elementAt(i)).x > 
			 ((MyPoint)pf.elementAt(j)).x) ) {
		       return false;
		   }
		}
	    }
	return true;
	}
    }

    /**
     *
     * Implementation of the run function from the Runnable interface.
     * Determines what the current step is and calls the appropriate method.
     *
     */
    public void run() {

	if (step_index_d == 1) {
	    disableControl();
	    step1();
	    enableControl();
	}
	if (step_index_d == 2) {
	    disableControl();
	    step2();
	    enableControl(); 
	}
	if (step_index_d == 3) {
	    disableControl();
	    step3(); 
	    enableControl(); 
	}
	if (step_index_d == 4) {
	    disableControl();
	    step4(); 
	    enableControl(); 
	}
	if (step_index_d == 5) {
	    disableControl();
	    step5(); 
	    enableControl(); 

	}
	return;
    }
    
    /**
     *
     * Displays data sets from input box in output box, clears the data
     * points from the vector for input data, zero-mean data,
     * does interpolation, and then finds mean of the interpolated data
     * interpolates the zero-mean data, displays the result
     *
     * @return   True
     *
     */
    boolean step1() {

	// set up progress bar
	//
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);

	if (set_1_d.size() > scale) {
	    scale = set_1_d.size();
	}

	// Display original data
	// 
	if(set_1_d.size() > 0 ) {
       	   interpolated_set_1_d.removeAllElements();
	   iset_1_d.removeAllElements();
	   mset1_d.removeAllElements();
	   yestimate_1_d.removeAllElements();
	   state_estimate_1_d.removeAllElements();
	   interpol(set_1_d, interpolated_set_1_d);
	   avg_1_d = mean(interpolated_set_1_d, mset1_d);
	   interpol(mset1_d, iset_1_d);
	   output_panel_d.addOutput(set_1_d, (Classify.PTYPE_INPUT), 
				    data_points_d.color_dset1);
	}

 	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	return true;
    }
	
   /**
    * find the LP coefficients using levinson durbin algorithm
    *
    * @return    true
    */
    boolean step2() {
	autoCorrelation();
	lpcCoefficient();

	pro_box_d.appendMessage(
			    new String(
			    " Class 0: \n\n" +
			    "    Variances \n"+
			    "     Q = process noise      = " + 
			    MathUtil.SetDecimal(covar_nse_pro_d, 4) + "\n" +
			    "     R = observation noise  = " + 
			    MathUtil.SetDecimal(covar_nse_obs_d,4) + "\n" +
			    "     P = estimate error     = " + 
			    MathUtil.SetDecimal(covar_nse_err_d,4) + "\n"));
	return true; 
    }   
    
    /**
     * First iteration of set_1_d, slow motion
     *
     * @return true
     */
    boolean step3() {

	
	int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(4);

	scale = iset_1_d.size();
	obsn_seq_d = new double[scale];

	for (int i = 0; i < scale; i++) {
	    obsn_seq_d[i] = iset_1_d.elementAt(i).y;
	}

	state_mem_d.initMatrixValue(lporder, particles_d, 
					0.0, Matrix.FULL);
	for ( ; idx < lporder; idx++) {
	    MyPoint p = new MyPoint();
	    p.y = ((MyPoint)iset_1_d.elementAt(idx)).y + avg_1_d;
	    p.x = ((MyPoint)iset_1_d.elementAt(idx)).x;
	    yestimate_1_d.addElement(new MyPoint(p));
	}

	for (int i = 0; i < lporder; i++) {
	    double value = -finallpc_1_d[lporder - i];
	    trn_d.setValue(lporder - 1, i, value);
	}
	
	// generate the first set of particles
	//
	Matrix mean_n_stddev = new Matrix();
	mean_n_stddev.initMatrixValue(lporder, 2, 0.0, Matrix.FULL);
	for (int i = 0; i < lporder; i++) {
	    double value = obsn_seq_d[i];
	    mean_n_stddev.setValue(i, 0, value);
	    mean_n_stddev.setValue(i, 1, Math.sqrt(covar_nse_err_d));
	}
	state_mem_d.gausRandGen(mean_n_stddev);    
	
	for (int i = 0; i < 1; i++) {
	// for (int i = idx; i < scale - 1; i++) {
	// for (int i = 0; i < scale - lporder; i++) {
   
	    // predict next state 
	    // x(k) = F * x(k) + covar_nse_pro_d
	    //
	    // declare a Matrix (lporder, 1)
	    // copy the column of state into it.
	    // multiply it by trn_d
	    // add covar_nse_pro_d to the last row
	    // copy back the result at that specific column
	    //
	    Vector<Double> proc_noise = new Vector<Double>();
	    proc_noise = gaussianGenerator(particles_d, 0, 
					   Math.sqrt(covar_nse_pro_d));
			
	    for (int j = 0; j < particles_d; j++) {

		// copy the specific column of state into another matrix
		//
		Matrix mat = new Matrix();
		mat.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		for ( int k = 0; k < lporder; k++) {
		    double val = state_mem_d.getValue(k, j);
		    mat.setValue(k, 0, val);
		}
		
		// get the next state for the above state particle
		//
		Matrix res = new Matrix();
		res.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		trn_d.multMatrix(mat, res);
		for ( int k = 0; k < lporder; k++) {
		    double val = res.getValue(k, 0);
		    state_mem_d.setValue(k, j, val);
		}
		
		// add process noise to the resulted state
		//
 		double value = state_mem_d.getValue(lporder - 1, j) + 
		    proc_noise.elementAt(j);
		state_mem_d.setValue(lporder - 1, j, value);
	    }

	    pro_box_d.setProgressCurr(++progress_curr);

	    try { Thread.sleep(500);
	    } catch (Exception e) {}
	    		
	    // Importance weights
	    // get the state_mem_d of interest (elements in the
	    // last row) <- this is the estimated obsn
	    //
	    double[] estimated_obsn = new double[particles_d];
	    for (int k = 0; k < particles_d; k++) {
		estimated_obsn[k] = state_mem_d.getValue(lporder - 1, k);
	    }

	    // get the current obsn
	    // 
	    double obsn = obsn_seq_d[idx];

	    // importance weights
	    //
	    Vector<Double> wts = new Vector<Double>();
	    double wts_sum = 0.0;
	
	    // denominator
	    //
	    for (int j = 0; j < particles_d; j++) {
		double temp = obsn - estimated_obsn[j];
		temp = temp * temp;
		temp = (-0.5 * temp) / covar_nse_obs_d;
		temp = Math.exp(temp);
		wts_sum = wts_sum + temp;
	    }	    
		
	    if (wts_sum < 0.05) {
		// assign all weights to be equal to 1/particles
		//
		double temp = 1.0 / particles_d;
		for( int k = 0; k < particles_d; k++) {
		    wts.addElement(temp);
		}
	    }
	    else {
		// divide each sum by the weight sum
		// 
		for (int k = 0; k < particles_d; k++) {
		    double temp = obsn - estimated_obsn[k];
		    temp *= temp;
		    temp *= -0.5 / covar_nse_obs_d;
		    temp = Math.exp(temp);
		    temp = temp / wts_sum;
		    wts.addElement(temp);
		}
	    }

	    // Update states
	    // for resample code:
	    // ref: http://www.cs.ubc.ca/~nando/software/demo1.tar 
	    //
	    Vector<Double> u = new Vector<Double>();
		
	    // used for cum sum of (weights) q
	    //
	    Vector<Double> Q = new Vector<Double>();
	    double cum_sum_q = 0;

	    // arbitrary normalized factor
	    //
	    Vector<Double> t = new Vector<Double>();
	    
	    // used for cumulative sum of t
	    //
	    Vector<Double> T = new Vector<Double>();
	    double cum_sum_t = 0;
		
	    for (int j = 0; j < (particles_d + 1); j++) {
		Random r = new Random();
		double temp = r.nextDouble(); 
		u.addElement(temp);
		temp = -1 * Math.log(temp);
		t.addElement(temp);
		cum_sum_t += temp;
		T.addElement(cum_sum_t);
	    }

	    double[] temp_state = new double[particles_d];
		
	    for (int j = 0; j < particles_d; j++) {
		temp_state[j] = avg_1_d;
		cum_sum_q += wts.elementAt(j);
		Q.addElement(cum_sum_q);
	    }
		
	    int j = 0;
	    int k = 0;
	    while (j < particles_d) {
		if ((Q.elementAt(j) * T.elementAt(particles_d - 1)) > 
		    T.elementAt(k)) {
		    temp_state[k] = state_mem_d.getValue(lporder - 1, j);
		    k++;
		}
		else { j++; }
	    }

	    for ( int a = 0; a < particles_d; a++) {
		state_mem_d.setValue(lporder - 1, a, temp_state[a]);
	    }

	    // get the mean of each state
	    //
	    double mean = 0;
	    for(int a = 0; a < particles_d; a++) 
		mean += temp_state[a];
	    mean = mean / particles_d;

	    // filtered data
	    //
	    MyPoint pt1 = new MyPoint(iset_1_d.elementAt(idx).x, 
				      mean + avg_1_d);
	    yestimate_1_d.addElement(pt1);

	    try { Thread.sleep(150);
	    }
	    catch (Exception e) {}

	    pro_box_d.setProgressCurr(++progress_curr);

	    try { Thread.sleep(150);
	    }
	    catch (Exception e) {}

	    // Generate distribution as points to display
	    //
	    Vector<MyPoint> prediction_points = new Vector<MyPoint>();
	    double curr_x_val = iset_1_d.elementAt(idx).x;

	    for (int r = 0; r < state_mem_d.getNumRows(); r++) {
		for (int c = 0; c < particles_d; c++) {

		    prediction_points.addElement(
		             new MyPoint(curr_x_val,
			         state_mem_d.getValue(r, c)));

		}
	    }

	    output_panel_d.addOutput(prediction_points,
				     (Classify.PTYPE_INPUT), Color.green);
	    output_panel_d.repaint();
		
	    pro_box_d.setProgressCurr(++progress_curr);
		
	    try { Thread.sleep(500);
	    } catch (Exception e) {}	    
	
	    idx++;
	    output_panel_d.removeOutput();
	    output_panel_d.addOutput(yestimate_1_d, Classify.PTYPE_LINE,
				     Color.blue);
	    output_panel_d.repaint();

	    // update progress bar
	    //
	    pro_box_d.setProgressCurr(++progress_curr);
	}
	    
	    pro_box_d.setProgressCurr(4);
	    return true;
    }

    /**
     * Second iteration of set_1_d, slow motion
     *
     * @return true
     */
    boolean step4() {

	int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(4);

	for (int i = 1; i < 2; i++) {
   
	    // predict next state 
	    // x(k) = F * x(k) + covar_nse_pro_d
	    //
	    // declare a Matrix (lporder, 1)
	    // copy the column of state into it.
	    // multiply it by trn_d
	    // add covar_nse_pro_d to the last row
	    // copy back the result at that specific column
	    //
	    Vector<Double> proc_noise = new Vector<Double>();
	    proc_noise = gaussianGenerator(particles_d, 0, 
					   Math.sqrt(covar_nse_pro_d));
			
	    for (int j = 0; j < particles_d; j++) {

		// copy the specific column of state into another matrix
		//
		Matrix mat = new Matrix();
		mat.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		for ( int k = 0; k < lporder; k++) {
		    double val = state_mem_d.getValue(k, j);
		    mat.setValue(k, 0, val);
		}
		
		// get the next state for the above state particle
		//
		Matrix res = new Matrix();
		res.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		trn_d.multMatrix(mat, res);
		for ( int k = 0; k < lporder; k++) {
		    double val = res.getValue(k, 0);
		    state_mem_d.setValue(k, j, val);
		}
		
		// add process noise to the resulted state
		//
 		double value = state_mem_d.getValue(lporder - 1, j) + 
		    proc_noise.elementAt(j);
		state_mem_d.setValue(lporder - 1, j, value);
	    }

	    pro_box_d.setProgressCurr(++progress_curr);

	    try { Thread.sleep(500);
	    } catch (Exception e) {}
	    		
	    // Importance weights
	    // get the state_mem_d of interest (elements in the
	    // last row) <- this is the estimated obsn
	    //
	    double[] estimated_obsn = new double[particles_d];
	    for (int k = 0; k < particles_d; k++) {
		estimated_obsn[k] = state_mem_d.getValue(lporder - 1, k);
	    }

	    // get the current obsn
	    // 
	    double obsn = obsn_seq_d[idx];

	    // importance weights
	    //
	    Vector<Double> wts = new Vector<Double>();
	    double wts_sum = 0.0;
	
	    // denominator
	    //
	    for (int j = 0; j < particles_d; j++) {
		double temp = obsn - estimated_obsn[j];
		temp = temp * temp;
		temp = (-0.5 * temp) / covar_nse_obs_d;
		temp = Math.exp(temp);
		wts_sum = wts_sum + temp;
	    }	    
		
	    if (wts_sum < 0.05) {
		// assign all weights to be equal to 1/particles
		//
		double temp = 1.0 / particles_d;
		for( int k = 0; k < particles_d; k++) {
		    wts.addElement(temp);
		}
	    }
	    else {
		// divide each sum by the weight sum
		// 
		for (int k = 0; k < particles_d; k++) {
		    double temp = obsn - estimated_obsn[k];
		    temp *= temp;
		    temp *= -0.5 / covar_nse_obs_d;
		    temp = Math.exp(temp);
		    temp = temp / wts_sum;
		    wts.addElement(temp);
		}
	    }

	    // Update states
	    // for resample code:
	    // ref: http://www.cs.ubc.ca/~nando/software/demo1.tar 
	    //
	    Vector<Double> u = new Vector<Double>();
		
	    // used for cum sum of (weights) q
	    //
	    Vector<Double> Q = new Vector<Double>();
	    double cum_sum_q = 0;

	    // arbitrary normalized factor
	    //
	    Vector<Double> t = new Vector<Double>();
	    
	    // used for cumulative sum of t
	    //
	    Vector<Double> T = new Vector<Double>();
	    double cum_sum_t = 0;
		
	    for (int j = 0; j < (particles_d + 1); j++) {
		Random r = new Random();
		double temp = r.nextDouble(); 
		u.addElement(temp);
		temp = -1 * Math.log(temp);
		t.addElement(temp);
		cum_sum_t += temp;
		T.addElement(cum_sum_t);
	    }

	    double[] temp_state = new double[particles_d];
		
	    for (int j = 0; j < particles_d; j++) {
		temp_state[j] = avg_1_d;
		cum_sum_q += wts.elementAt(j);
		Q.addElement(cum_sum_q);
	    }
		
	    int j = 0;
	    int k = 0;
	    while (j < particles_d) {
		if ((Q.elementAt(j) * T.elementAt(particles_d - 1)) > 
		    T.elementAt(k)) {
		    temp_state[k] = state_mem_d.getValue(lporder - 1, j);
		    k++;
		}
		else { j++; }
	    }

	    for ( int a = 0; a < particles_d; a++) {
		state_mem_d.setValue(lporder - 1, a, temp_state[a]);
	    }

	    // get the mean of each state
	    //
	    double mean = 0;
	    for(int a = 0; a < particles_d; a++) 
		mean += temp_state[a];
	    mean = mean / particles_d;

	    // filtered data
	    //
	    MyPoint pt1 = new MyPoint(iset_1_d.elementAt(idx).x, 
				      mean + avg_1_d);
	    yestimate_1_d.addElement(pt1);

	    try { Thread.sleep(150);
	    }
	    catch (Exception e) {}

	    pro_box_d.setProgressCurr(++progress_curr);

	    try { Thread.sleep(150);
	    }
	    catch (Exception e) {}

	    // Generate distribution as points to display
	    //
	    Vector<MyPoint> prediction_points = new Vector<MyPoint>();
	    double curr_x_val = iset_1_d.elementAt(idx).x;

	    for (int r = 0; r < state_mem_d.getNumRows(); r++) {
		for (int c = 0; c < particles_d; c++) {

		    prediction_points.addElement(
		             new MyPoint(curr_x_val,
			         state_mem_d.getValue(r, c)));
		}
	    }

	    output_panel_d.addOutput(prediction_points,
				     (Classify.PTYPE_INPUT), Color.green);
	    output_panel_d.repaint();
		
	    try { Thread.sleep(250);
	    }
	    catch (Exception e) {}
	    pro_box_d.setProgressCurr(++progress_curr);
		
	    try { Thread.sleep(250);
	    } catch (Exception e) {}	    
	
	    idx++;
	    output_panel_d.removeOutput();
	    output_panel_d.addOutput(yestimate_1_d, Classify.PTYPE_LINE,
				     Color.blue);
	    output_panel_d.repaint();

	    // update progress bar
	    //
	    pro_box_d.setProgressCurr(++progress_curr);
	    
	}
 	pro_box_d.setProgressCurr(4);
 	return true;
    }

    /**
     * Steps through the particle filtering algorithm
     *
     * @return    true
     *
     */
    boolean step5() {

	int progress_curr = 0;
	int progress_max = interpolated_set_1_d.size(); 
	pro_box_d.setProgressMax(scale - lporder - 1);
	pro_box_d.setProgressCurr(0);
	
	for (int i = 2; i < scale - lporder; i++) {
   
	    // predict next state 
	    // x(k) = F * x(k) + covar_nse_pro_d
	    //
	    // declare a Matrix (lporder, 1)
	    // copy the column of state into it.
	    // multiply it by trn_d
	    // add covar_nse_pro_d to the last row
	    // copy back the result at that specific column
	    //
	    Vector<Double> proc_noise = new Vector<Double>();
	    proc_noise = gaussianGenerator(particles_d, 0, 
					   Math.sqrt(covar_nse_pro_d));
			
	    for (int j = 0; j < particles_d; j++) {

		// copy the specific column of state into another matrix
		//
		Matrix mat = new Matrix();
		mat.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		for ( int k = 0; k < lporder; k++) {
		    double val = state_mem_d.getValue(k, j);
		    mat.setValue(k, 0, val);
		}
		
		// get the next state for the above state particle
		//
		Matrix res = new Matrix();
		res.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		trn_d.multMatrix(mat, res);
		for ( int k = 0; k < lporder; k++) {
		    double val = res.getValue(k, 0);
		    state_mem_d.setValue(k, j, val);
		}
		
		// add process noise to the resulted state
		//
 		double value = state_mem_d.getValue(lporder - 1, j) + 
		    proc_noise.elementAt(j);
		state_mem_d.setValue(lporder - 1, j, value);
	    }

	    try { Thread.sleep(100);
	    } catch (Exception e) {}
	    		
	    // Importance weights
	    // get the state_mem_d of interest (elements in the
	    // last row) <- this is the estimated obsn
	    //
	    double[] estimated_obsn = new double[particles_d];
	    for (int k = 0; k < particles_d; k++) {
		estimated_obsn[k] = state_mem_d.getValue(lporder - 1, k);
	    }

	    // get the current obsn
	    // 
	    double obsn = obsn_seq_d[idx];

	    // importance weights
	    //
	    Vector<Double> wts = new Vector<Double>();
	    double wts_sum = 0.0;
	
	    // denominator
	    //
	    for (int j = 0; j < particles_d; j++) {
		double temp = obsn - estimated_obsn[j];
		temp = temp * temp;
		temp = (-0.5 * temp) / covar_nse_obs_d;
		temp = Math.exp(temp);
		wts_sum = wts_sum + temp;
	    }	    
		
	    if (wts_sum < 0.05) {
		// assign all weights to be equal to 1/particles
		//
		double temp = 1.0 / particles_d;
		for( int k = 0; k < particles_d; k++) {
		    wts.addElement(temp);
		}
	    }
	    else {
		// divide each sum by the weight sum
		// 
		for (int k = 0; k < particles_d; k++) {
		    double temp = obsn - estimated_obsn[k];
		    temp *= temp;
		    temp *= -0.5 / covar_nse_obs_d;
		    temp = Math.exp(temp);
		    temp = temp / wts_sum;
		    wts.addElement(temp);
		}
	    }

	    // Update states
	    // for resample code:
	    // ref: http://www.cs.ubc.ca/~nando/software/demo1.tar 
	    //
	    Vector<Double> u = new Vector<Double>();
		
	    // used for cum sum of (weights) q
	    //
	    Vector<Double> Q = new Vector<Double>();
	    double cum_sum_q = 0;

	    // arbitrary normalized factor
	    //
	    Vector<Double> t = new Vector<Double>();
	    
	    // used for cumulative sum of t
	    //
	    Vector<Double> T = new Vector<Double>();
	    double cum_sum_t = 0;
		
	    for (int j = 0; j < (particles_d + 1); j++) {
		Random r = new Random();
		double temp = r.nextDouble(); 
		u.addElement(temp);
		temp = -1 * Math.log(temp);
		t.addElement(temp);
		cum_sum_t += temp;
		T.addElement(cum_sum_t);
	    }

	    double[] temp_state = new double[particles_d];
		
	    for (int j = 0; j < particles_d; j++) {
		temp_state[j] = avg_1_d;
		cum_sum_q += wts.elementAt(j);
		Q.addElement(cum_sum_q);
	    }
		
	    int j = 0;
	    int k = 0;
	    while (j < particles_d) {
		if ((Q.elementAt(j) * T.elementAt(particles_d - 1)) > 
		    T.elementAt(k)) {
		    temp_state[k] = state_mem_d.getValue(lporder - 1, j);
		    k++;
		}
		else { j++; }
	    }

	    for ( int a = 0; a < particles_d; a++) {
		state_mem_d.setValue(lporder - 1, a, temp_state[a]);
	    }

	    // get the mean of each state
	    //
	    double mean = 0;
	    for(int a = 0; a < particles_d; a++) 
		mean += temp_state[a];
	    mean = mean / particles_d;

	    // filtered data
	    //
	    MyPoint pt1 = new MyPoint(iset_1_d.elementAt(idx).x, 
				      mean + avg_1_d);
	    yestimate_1_d.addElement(pt1);

	    try { Thread.sleep(100);
	    }
	    catch (Exception e) {}


	    // Generate distribution as points to display
	    //
	    Vector<MyPoint> prediction_points = new Vector<MyPoint>();
	    double curr_x_val = iset_1_d.elementAt(idx).x;

	    for (int r = 0; r < state_mem_d.getNumRows(); r++) {
		for (int c = 0; c < particles_d; c++) {

		    prediction_points.addElement(
		             new MyPoint(curr_x_val,
			         state_mem_d.getValue(r, c)));
		}
	    }
		

	    output_panel_d.addOutput(prediction_points,
				     (Classify.PTYPE_INPUT), Color.green);
	    output_panel_d.repaint();
		
	    try { Thread.sleep(200);
	    } catch (Exception e) {}	    
	
	    idx++;
	    output_panel_d.removeOutput();
	    output_panel_d.addOutput(yestimate_1_d, Classify.PTYPE_LINE,
				     Color.blue);
	    output_panel_d.repaint();

	    // update progress bar
	    //
	    pro_box_d.setProgressCurr(++progress_curr);
	    
	}
	pro_box_d.setProgressCurr(progress_max);
	output_panel_d.repaint();
       double compute_err = 
	   actual_error( yestimate_1_d, interpolated_set_1_d);
	    pro_box_d.appendMessage(" Class 0: \n" + 
				    "  Mean Square Error   = " +
				 MathUtil.SetDecimal(compute_err, 6) + "\n");

	pro_box_d.appendMessage("Algorithm Complete" + "\n");	
	pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);


	return true;
    }
 
    /**
     * This method generates the gaussian distribution 
     *
     * @param num_samples the number of gaussian points to be generated
     * @param mean        the average value of the distribution
     * @param std_dev     the standard deviation
     *
     * @return            the vector containing the gaussian distribution
     */
    Vector<Double> gaussianGenerator(int num_samples,
				     double mean, 
				     double std_dev) {
	Vector<Double> gauss_vec = new Vector<Double> ();
	Random r = new Random();
	for (int i = 0; i < num_samples; i++) 
	    gauss_vec.addElement((r.nextGaussian() * std_dev) + mean);

	return gauss_vec;
    }

    /**
     *
     * Calculates the interpolated points for the data inputs   
     *
     * @param  v  input data points
     * @param  iset  interpolated data points
     *
     */
    public void interpol(Vector<MyPoint> v , Vector<MyPoint> iset)
    {	
	double sample;
	double d ;
	MyPoint r = new MyPoint();	
	
	double[] y2 = new double[v.size()];  
	double[] y  = new double[v.size()];
	double[] x  = new double[v.size()];
       
       for (int k = 0; k < v.size(); k++)
       {
	   x[k] = ((MyPoint)v.elementAt(k)).x;
	   y[k] = ((MyPoint)v.elementAt(k)).y;
       }

       // Reference:
       // "Numerical Recipe in C", Cambridge University Press, pp 113-116
       
       // total time interval of the data points calculated
       //
       d = ((MyPoint)v.lastElement()).x - ((MyPoint)v.firstElement()).x ; 
       
       // time step calculation
       //
       sample = d / scale;
       spline(x, y, y2, v.size());
       
       // iset holds new data points - interpolated values
       // First value added 
       //
       iset.addElement(new MyPoint((MyPoint)v.firstElement()));
       
       // Increment to next Time value
       //
      	r.x = sample + ((MyPoint)v.firstElement()).x;

	while (r.x <=  ((MyPoint)v.lastElement()).x)
	{
	    for(int j = 1; j < v.size(); j++)
	    {
		// the New Interpolated amplitude will be calculated
		// for the two points within which the time intervals 
		// lies, Hence, the condition check
		//
		if((r.x <= ((MyPoint)v.elementAt(j)).x ) && 
		   (r.x >=  ((MyPoint)v.elementAt(j - 1)).x)) 
		{
		    r.y = 0 ;
		    splint ((MyPoint)v.elementAt(j - 1), 
			    (MyPoint)v.elementAt(j), r, y2, (j - 1));

		    // New value appended to the array
		    //
		    iset.addElement(new MyPoint(r));
		    break; 
		}
	    }	    
   	    r.x = r.x + sample;
	}
	// sometimes the last element of the user entered 
	// values is missed out hence the condition to 
	// cross-check and put it in
	//
	if (((MyPoint)iset.lastElement()).x != ((MyPoint)v.lastElement()).x )
	    iset.addElement( new MyPoint((MyPoint)v.lastElement()));
    }

    /**
     *
     * Actually interpolates the points
     *
     * @param x    array containing the x coordinates of datapoints
     * @param y    array containing the y coordinates of datapoints
     * @param y2   array containing the interpolated y coordinates
     * @param size the size of the array to be interpolated
     * 
     */
    public void spline (double[] x, double[] y, double[] y2, int size )
    { 
	double[] u = new double[size];
	double p; 
	double sig; 
	
	y2[0] = 0;
	u[0] = 0;
	 
	for( int i = 1; i < size - 1; i++ )
	{ 
	    sig = ( x[i] - x[i - 1] ) / ( x[i + 1] - x[i - 1] ) ;
	    p = sig * y2[i - 1] + 2 ;
	    y2[i] = ( sig - 1) / p ;
	    u[i] = ( y[i + 1] - y[i] ) / ( x[i + 1] - x[i] ) ;
	    u[i] = ( 6 * u[i] / ( x[i + 1] - x[i - 1] ) 
		     - sig * u[i - 1] ) / p ;
	}
	y2[size - 1] = 0 ;
	
	// This is the back substitution loop of the tridiagonal algorithm
	//
	for(int i = size - 2; i >= 0; i-- )     
	    y2[i] = y2[i] * y2[i + 1] + u[i] ;
    }

    /**
     * Interpolates for a point between the two known points
     * using Cubic Interpolation
     *
     * @param   u1 start point for the interpolation
     * @param   u2 end point for the interpolation
     * @param   r  returning point, basically the interpolated point
     * @param   y2 array used for reassigning of r
     * @param   i  the sample number
     *
     */
    public void splint ( MyPoint u1, MyPoint u2, MyPoint r, 
			 double[] y2, int i )
    {
	double h;
	double a;
	double b;
 
	h = u2.x - u1.x ;
	a = ( u2.x - r.x ) / h ;
	b = ( r.x - u1.x ) / h ;
	r.y = a * u1.y + b * u2.y + ((a * a * a - a) * y2[i] 
	                             + ( b * b * b - b ) * y2[i + 1] ) 
	                                                 * ( h * h ) / 6 ; 
    }

    /**
     *
     * Calculates the mean and the zero-mean data points  
     *
     * @param    v orginal datapoints
     * @param    mv zero mean datapoints 
     *
     * @return   The average in a double
     * 
     */
    public double mean(Vector<MyPoint> v, Vector<MyPoint> mv)
    {
        double average = 0;
	MyPoint p      = new MyPoint();

       for(int i = 0; i < v.size(); i++ )
	   average = average + ((MyPoint)v.elementAt(i)).y;
       average = average / v.size();
       
       for(int i = 0; i < v.size(); i++ )
       {
	   p.y = ((MyPoint)v.elementAt(i)).y - average;
	   p.x =  ((MyPoint)v.elementAt(i)).x; 
	   mv.addElement(new MyPoint(p));
       }        
       return average ;
    }  
   
    /**
    *
    * Computes the autocorrelation coeffient from the data sets   
    *
    */
    public void autoCorrelation()
    {
	if (iset_1_d.size() > 0 ) {
	    autoc_1_d = new double[lporder + 1];
	    autocorrelate(iset_1_d, autoc_1_d);
	}	
    }

    /**
     *
     * Actaully computes the autocorrelation coefficients
     *
     * @param  v Vector of datapoints
     * @param  autoCoeff_co array of autocorrelation coefficients
     *
     */
    public void autocorrelate ( Vector v, double[] autoCoeff_co) 
    {
	int size;
	size                 = v.size();
      	double[] amplitude_y = new double[size];

      	// to store the amplitude for calculation autocoefficient
      	//
    	for (int i = 0; i < size; i++)
	    amplitude_y[i] = ((MyPoint)v.elementAt(i)).y;

      	for (int j = 0; j <= lporder; j++)
      	{
      	    autoCoeff_co[j] = 0;
      	    if ( j < size)
      		for (int k = 0; k < size - j; k++)
		    autoCoeff_co[j] = autoCoeff_co[j] 
			            + amplitude_y[k] * amplitude_y[k + j];
	}

   	// normalize the autocorrelation cofficient
      	//
     	for (int i = lporder; i >= 0; i--) 
      	    autoCoeff_co[i] = autoCoeff_co[i] / autoCoeff_co[0];	  
    }

    /**
     *
     * Computes the Linear Prediction coefficient from the data sets   
     *
     */
    public void lpcCoefficient()
    {
	
	// Rabiner, Schafer, "Digital Processing of Speech signals", 
	// Bell Laboratories, Inc. 1978, 
	// section 8.3.2 Durbin's recursive solution for 
	// the autocorrelation equations pp 411-412
	//	
	if (set_1_d.size() > 0)
	{
	    finallpc_1_d = new double [lporder + 1];
	    ref_coeff_1_d = new double [lporder + 1];
	    
	}
    }

    /**
     *
     * Actually calculate the LP coefficient and the Residual Error 
     * Energy, and Reflection Coefficients
     *
     * @param  auto_coeff array of auto correlation coefficients
     * @param  lpc        array of linear prediction coefficients
     * @param  rc_reg     array of reflection coefficients
     * 
     * @return residual error energy in a double
     *
     */
    public double calculate_lpc (double[] auto_coeff, 
				 double[] lpc, double[] rc_reg)
    {
	int j_bckwd       = 0;
	long middle_index = 0;

	// Reference:
	// J.D. Markel and A.H. Gray, "Linear Prediction of Speech,"
	// Springer-Verlag Berlin Heidelberg, New York, USA, pp. 219,
	// 1980.
 
	// initialization
	//
	lpc[0] = 1;
	double err_energy = auto_coeff[0];

	// if the energy of the signal is zero we set all the predictor
	// coefficients to zero and exit
	//
	if (err_energy == (double)0)
	{
	    for (int i = 0; i < lporder + 1 ; i++)
		lpc[i] = 0;
	}
	else
        {
	    // do the first step manually
	    //
	    double sum = 0;
	    double tmp = 0;
	    
	    sum         = -auto_coeff[1] / err_energy;
	    rc_reg[1]   = sum;
	    lpc[1]      = rc_reg[1];
	    tmp         = 1 - rc_reg[1] * rc_reg[1];
	    err_energy *= tmp;
	    // recursion
	    //
	    for (int i = 2; i <= lporder; i++)
	    {
		sum = 0;
		for (int j = 1; j < i; j++)
		    sum += lpc[j] * auto_coeff[i - j];
 
		rc_reg[i]    = -(auto_coeff[i] + sum) / err_energy;	    
		lpc[i]       = rc_reg[i];
		j_bckwd      = i - 1;
		middle_index = i / 2;
		for (int j = 1; j <= middle_index; j++)
		{
		    sum          = lpc[j_bckwd] + rc_reg[i] * lpc[j];
		    lpc[j]       = lpc[j] + rc_reg[i] * lpc[i - j];
		    lpc[j_bckwd] = sum;
		    j_bckwd--;
		}
		// compute new error
		//
		tmp = 1.0 - rc_reg[i] * rc_reg[i];
		err_energy *= tmp;
	    }
	}
	return err_energy;
    }

    /**
     *
     * Calculates the estimated points for the data inputs
     *
     */
    public void final_estimate()
    {	
	if(iset_1_d.size() > 0)
	    estimate (iset_1_d, yestimate_1_d, avg_1_d, finallpc_1_d);
    } 
    
    /**
     *
     * Estimates the amplitude based on the LP coeficients.
     *
     * @param    iset         interpolated data points
     * @param    y_estimate   predicted final signal data points
     * @param    avg          mean of the original datapoints given
     * @param    final_lpc    array of final linear prediction coefficients
     * 
     */
    public void estimate( Vector<MyPoint> iset, Vector<MyPoint> y_estimate, 
			  double avg, double[] final_lpc)
    {
	   y_estimate.removeAllElements();
	   double amplitude[];
	
	   for(int i = 0; i < iset.size(); i++)
	       y_estimate.addElement(new MyPoint((MyPoint)iset.elementAt(i))); 

	   amplitude = new double[y_estimate.size()];
	   
	   for (int i = 0; i < y_estimate.size(); i++)
	       amplitude[i] = ((MyPoint)y_estimate.elementAt(i)).y;
	   
	   ((MyPoint)y_estimate.firstElement()).y = 0 ;
	   
	   for ( int i = 1; i < y_estimate.size(); i++ )
	   {    
	       int z = i;
  	       double sum  = 0;
	       
	       for( int j = 1; (j <= lporder) && (z > 0); j++, z-- )
		   sum = sum - amplitude[z - 1] * final_lpc[j];
	       ((MyPoint)y_estimate.elementAt(i)).y = sum + avg;
	    }
    }

    /**
     *
     * Compute the actual error from the given data points and the estimated
     * values.
     *
     * @param  y_estimate datapoints of the estimated datapoints
     * @param  iset       original datapoints 
     *
     * @return actual error energy in a double
     */
    public double actual_error (Vector y_estimate, Vector iset)
    {
	double error_value;
	double act_error = (double) 0;
	int j = 0;
	int i = 0;

	// first point and the last point have their x-coordinates same
	// So the difference in y-values is the error value
	//
	error_value = ((MyPoint)y_estimate.firstElement()).y 
	              - ((MyPoint)iset.firstElement()).y;
	act_error = act_error + error_value * error_value; 
	error_value = ((MyPoint)y_estimate.lastElement()).y 
	              - ((MyPoint)iset.lastElement()).y;
	act_error = act_error + error_value * error_value;

	// for next values
	// need to continue till all the user defined points are exhausted
	//
	j++;

	for (i = 1; ( (i < y_estimate.size()) && (j < iset.size()) ); i++)
	{ 
	    while ( (((MyPoint)y_estimate.elementAt(i)).x < 
		     ((MyPoint)iset.elementAt(j)).x) 
		    && ((i < y_estimate.size() - 1) && (j < iset.size())) )
	    {
		    i++;
	    }

	    if ( j < iset.size())
	    {
		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     == ((MyPoint)iset.elementAt(j)).x ) 
		{
		    error_value = ((MyPoint)y_estimate.elementAt(i)).y 
			          - ((MyPoint)iset.elementAt(j)).y;
		    act_error = act_error + error_value * error_value;
		}

		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     > ((MyPoint)iset.elementAt(j)).x )
		{
		    double y1 = ((MyPoint)y_estimate.elementAt(i)).y;
		    double y2 = ((MyPoint)y_estimate.elementAt(i - 1)).y;
		    double x1 = ((MyPoint)y_estimate.elementAt(i)).x;
		    double x2 = ((MyPoint)y_estimate.elementAt(i - 1)).x;
		    double x_unknown = ((MyPoint)iset.elementAt(j)).x;
		    error_value = y2 - ( (x2 - x_unknown) * 
					 ( y2 - y1) / (x2 - x1));
		    act_error = act_error + error_value * error_value ;
		}
	    }
	    j++;
	}
	return act_error;
    }
} //end of class
