 /*
 * AlgorithmUKF.java v6.2 
 * Created by: Sanjay Patil 01/19/2006
 * 
 * Description: Unscented Kalman Filtering
 * the unscented Kalman filter follows:
 *
 * Reference:
 * R. van der Merve, N. de Freitas, A. Doucet, and E. Wan, 
 * "The Unscented Particle Filter," Technical Report CUED/F-INFENG/TR
 * 380, Cambridge University Engineering Department, Cambridge
 * University, U.K., August 2000 
 *
 *  E. Wan and R. Merwe, "The Unscented Kalman Filter for Nonlinear
 *  Estimation," Adaptive systems for Signal Processing,
 *  Communications, and Control Symposium 2000, AS-SPCC, pp 153-158.
 *  
 * for the unscented Kalman filter (UKF), the given starting values are
 * converted into sigma points and a similar procedure to the
 * kalman filter is used. UKF are generally used with nonlinear
 * Gaussian state space models
 *
 * kalman filters are used to model signals using sequential prediction
 * techniques. a kalman filter is represented by two state space equations:
 *
 *  state equation :       x(k) = A x(k-1) + u(k-1)
 *  observation equation:  y(k) = H x(k)   + v(k)
 *  using basic matrix computation.
 *
 * where:
 *  k -> time instant
 *  x -> hidden (to find) state vector
 *  y -> observation (measurement) vector
 *  u -> process noise - Gaussian N(0, Q)
 *  v -> observation noise - Gaussian N(0, R)
 *  A -> state transition model (matrix)
 *  H -> observation model (matrix)
 *  P -> estimate error covariance (mentioned below)
 *
 * the following parameters are given or known:
 *
 *  A, H, u(k), v(k), x(0), P(0), y(k)
 * 
 * kalman filtering consists of two steps:
 *
 * stage 1:  prediction stage.
 *
 *  stage 1a: predict the state based on the previous state.
 *            x^(k) = A x(k-1)
 *
 *  stage 1b: predict the initial estimate error
 *            P^(k) = A P(k-1) A'
 *
 * stage 2:  Correction stage (accept the observation)
 *
 *  stage 2a: find Kalman gain
 *            K(k) = P^(k) H' [H P^(k) H' + R]inverse
 *
 *  stage 2b: correct the state
 *            x(k) = x^(k) + K(k) [ y(k) - H x^(k) ]
 *
 *  stage 2c: correct the estimate error
 *            P(k) = [ I - K(k) H] P^(k)
 *
 * where,
 *        ^ denotes predicted values,
 *        ' denotes transpose,
 *        I denotes identity matrix of appropriate dimension
 * the hidden states are estimated using the observables.
 * the next state is predicted based on the previous state and transition
 * model. a degree of uncentainty associated with the estimation is
 * accounted for by process noise. based on the above predicted state,
 * the observation is predicted. the difference between the predicted
 * observation and the actual observation is the error. a gain
 * term corrects the predicted state by a term proportional to the error.
 *
 *
 * The algorithm goes as follows
 *
 * 1. set the initial state vector, augument it with process noise
 *    observation noise means.  
 *    Repeat the above exercise for state covariance matrix so, if
 *    state vector is of length 4, and each of noise means are of
 *    length * 1, augumented state vector will be of length
 *    6. Similarly, state * covariance matrix (will be of dimensions 4
 *    x 4) after augumentation * will be 6 x 6. number of sigma points
 *    will be 2 * aug_len + 1.  * assume values for alpha, beta, and
 *    kappa.  1. generate sigma points * based on augumented state
 *    vector and augumented state covariance * matrix (using augumente
 *    length and lambda)
 *
 * 2. time update equations
 * 2a. find the state update using transition matrix and process noise
 *     X_x(k) = F ( X_x(k-1), X_pro_nse(k-1)) 
 * 2b. evaluate mean state values based on weights w_m and X_x(k)
       X_x_mean(k) = summation [w_m * X_x(k)]
 * 2c. compute state covariance using weights w_c and state sigma
 *     and state mean
 *     covar_nse_pro(k) = summation [ w_c * E{diff(X_x_mean, X_x)}]
 *     E{diff(X_x_mean, X_x)} = diff(X_x_mean, X_x) * diff(X_x_mean, X_x)'
 * 2d. compute obsn sigma points using state sigma and obsn noise sigma
 *     X_y(k) = H (X_x(k), X_obsn_nse(k-1))
 * 2e. compute obsn mean based on weight w_m and obsn sigma points
 *     X_y_mean(k) = summation [ w_m * X_y]
 *
 * 3. measurement update equations
 * 3a. compute obsn covariance based on weights w_c, obsn sigmas, obsn means
 *     covar_nse_obs(k) = summation [ w_c * E{diff(X_y, X_y_mean)}]
 *     E{diff(X_y, X_y_mean)} = diff(X_y, X_y_mean) * diff(X_y, X_y_mean)'
 * 3b. compute cross covariance based on w_c, obsn sigmas, obsn means,
 *     state sigmas, state means
 *     covar_nse_xcr(k) = summation [ w_c * diff(X_x_mean, X_x) 
 *                                        * diff(X_y, X_y_mean) ]
 * 3c. compute kalman gain based on covar_nse_obs and covar_nse_xcr
 *     kgain(k) = covar_nse_obs * covar_nse_xcr
 * 3d. compute corrected state vector based on kalman gain, actual
 *     obsn, estimated obsn (obsn mean) and state mean
 *     x(k) = x(k) + kgain(k) * (actual_obsn - X_y_mean);
 * 3e. correct state covariance using kalman gain, covar_nse_obs
       state_covar = state_covar - kgain(k) * covar_nse_obs * kgain(k)
 *
 */


//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

/**
 * Class that handles the Unscented Kalman filter algorithm
 */
public class AlgorithmUKF extends Algorithm {

    //-----------------------------------------------------------------
    //
    // static data members, for interpolation
    //
    //-----------------------------------------------------------------
    int lporder;
    int iporder;
    int scale;

    //-----------------------------------------------------------------
    //
    // primitive data members
    //
    //-----------------------------------------------------------------
    int output_canvas_d[][];   

    //-----------------------------------------------------------------
    //
    // instance data members
    //
    //-----------------------------------------------------------------
    String algo_id = "AlgorithmUKF"; 

    /**
     * Original data
     */
    Vector<MyPoint> set_1_d = new Vector<MyPoint>(40, 20);

    /**
     * Stores state estimate results from KF algorithm
     */
    Vector<MyPoint> state_estimate_1_d = new Vector<MyPoint>();
 
    // Interpolation tools -------
    //
    /**
     * for interpolation function, holds original points
     */
    Vector<MyPoint> iset_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * Vector for mean-subtracted [zero-mean] data meaning the original data
     * is interpolated and the average y-value of the set is zero
     */
    Vector<MyPoint> mset_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * For display purpose meaning it contains original data with points
     * in between
     */
    Vector<MyPoint> interpolated_set_1_d = new Vector<MyPoint>(40, 20); 
 
    /**
     * Interpolation tool, for algo PF and KF
     */
    Vector<MyPoint> yestimate_1_d = new Vector<MyPoint>(40, 20); 

    /**
     * to store the average values for each class
     */
    double avg_1_d;

    /**
     * to store autoCorrelation coefficient
     */
    double autoc_1_d[];

    /**
     * to store LP coefficient
     */
    double finallpc_1_d[]; 

    /**
     * to store residual energy 
     */
    double est_err_1_d = 0;	

    /**
     * to store actual error energy
     */
    double actual_err_1_d = 0;

    /**
     * to store the reflection coefficients
     */
    double ref_coeff_1_d[];

    // end interpolation tools ------
    //

    /**
     * kalman gain for each set
     */
    Matrix k_gain_d = new Matrix();
 
     /**
     * observation transition matrix
     */
    Matrix obs_d = new Matrix();

    /**
     * state transition matrix 
     */
    Matrix trn_d = new Matrix();

    /**
     * Process noise variance, Q
     */
    double covar_nse_pro_d;

    /**
     * Measurement noise variance, R
     */
    double covar_nse_obs_d;

    /**
     * Estimate Error variance, P
     */
    Matrix covar_nse_err_d = new Matrix(); 

    /** 
     * initial state estimate
     */
    Matrix state_mem_d = new Matrix();

    /** 
     * observation sequence 
     * will store the interpolated signal values 
     */
    double obs_seq_d[];
 
    /**
     * filtered data (will store the state means from sigmas)
     */
    double filt_data_d[];

    /** 
     * define constant parameters
     * alpha, beta, kappa, lambda, wts_m, wts_c
     */ 
    double alpha_d, beta_d, kappa_d, lambda_d;
    double wts_m[];
    double wts_c[];

    /**
     * index for the observation, length of augumented state vector
     */
    int idx, aug_len, sigma_len;

    //---------------------------------------------------------------
    //
    // class methods
    //
    //---------------------------------------------------------------

    /**
     * 
     * Implements the initialize() method in the base class. Initializes
     * member data and prepares for execution of first step. This method
     * "resets" the algorithm.
     *
     * @return   returns true if sets of data are valid
     *
     */
    public boolean initialize() {

		point_means_d           = new Vector<MyPoint>();
	description_d           = new Vector<String>();

	yestimate_1_d = new Vector<MyPoint>();
	state_estimate_1_d = new Vector<MyPoint>();

	step_count = 5;
	step_index_d = 0;

	// accepting the LP order and interpolation order
	//
	lporder    = Classify.main_menu_d.lporder;
	iporder    = Classify.main_menu_d.iporder;
	scale      = lporder * iporder;
	aug_len = lporder + 2;
	sigma_len = 2 * aug_len + 1;

	// configuring the transition matrix and observation matrix
	//
	trn_d.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	if (lporder == 1) {
	    trn_d.setValue(0, 0, Classify.main_menu_d.trn);
	}
	else {
	    for (int i = 0; i < lporder - 1; i++) {
		int j = i + 1;
		trn_d.setValue(i, j, 1.0);
	    }
	    trn_d.setRow(lporder - 1, 1);
	}

	obs_d.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
	if (lporder == 1) {
	    obs_d.setValue(0, lporder - 1, Classify.main_menu_d.lporder);
	}
	else {
	    obs_d.setValue(0, lporder - 1, 1);
	}
	
	// configure / set noise parameters
	// this should be the actual code
	//
	covar_nse_err_d.initMatrixValue(aug_len, aug_len, 0.0, Matrix.FULL);

	covar_nse_err_d.setDiagonal(Classify.main_menu_d.covar_nse_err);
	covar_nse_err_d.setValue(lporder, lporder, 
				 Classify.main_menu_d.covar_nse_pro);
	covar_nse_err_d.setValue(lporder + 1, lporder + 1, 
				 Classify.main_menu_d.covar_nse_obs);

	// defualt values to check the operation
	//
// 	// value taken from /cavs/hse/ies/isip/users/patil/asgmts/
// 	// asgmt_kf/kfukfpf/ukf/ukf_ar/MainUnscented.m
// 	covar_nse_err_d.setDiagonal(0.001);
// 	covar_nse_err_d.setValue(lporder, lporder, 0.03 );
// 	covar_nse_err_d.setValue(lporder + 1, lporder + 1, 0.02398);
// 	alpha_d  = 0.001;
// 	beta_d   = 2.0;
// 	kappa_d  = 0.0;

	// set UKF parameters
	//
	alpha_d  = Classify.main_menu_d.alpha;
	beta_d   = Classify.main_menu_d.beta;
	kappa_d  = Classify.main_menu_d.kappa;
	lambda_d = alpha_d * alpha_d * (aug_len + kappa_d) - aug_len;

	// kalman gain parameter dimensions set
	//
	k_gain_d.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);

	// index for the observation sequence initialized
	idx = 0;
	
	// Add the process description for the KF algorithm
	//
	if (description_d.size() == 0) {

	    String str = new String("  0.  Checking for the Data Validity");
	    description_d.addElement(str);
	    
	    str = new String("  1.  Displaying the Input Data Points");
	    description_d.addElement(str);
	    
	    str = new String("  2.  The recursive nature of \n" + 
			     "       unscented kalman filtering \n" +
			     "       predics states point by point \n" +
			     "       as shown in next step. \n" +
			     "\n" +
			     "  State Equation:\n" +
			     "    X(k) = (F) * X(k-1) + v(k) \n"+
			     "  Observation Equation:\n" +
			     "    Y(k) = (M) * X(k) + n(k) \n" );
	    description_d.addElement(str);
	    
	    str = new String("  3.  First iteration");
	    description_d.addElement(str);

	    str = new String("  4.  Second iteration");
	    description_d.addElement(str);

	    str = new String("  5.  Remaining iterations \n");
	    description_d.addElement(str);
	}
	set_1_d = data_points_d.dset1;

	if (checkdata_UKF(set_1_d) == true) {
	    pro_box_d.appendMessage((String)description_d.get(step_index_d));
	    pro_box_d.appendMessage("         Data points valid");
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    return true;	
	}
	else {
	    pro_box_d.appendMessage("\n" + "Invalid data");
	    pro_box_d.appendMessage("Clear and enter valid data" + "\n");
	    pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	    return false;
	}
    }

    /**
    * 
    * Validates the class entered by user for unscented Kalman Filtering  
    *
    * @param    ukf
    * @return   true if data is Vector kf is valid. It is invalid
    *           if the size of ukf is 1 or points are not entered
    *           in time progression
    * 
    */
    public boolean checkdata_UKF(Vector<MyPoint> ukf) {

	if ( ukf.size() <= 1) {
	    return false;
	}
	else {

	    for(int i = 0; i <= ukf.size() - 1 ; i++ ) {
		for(int j = i + 1; j <= ukf.size() - 1 ; j++ ) {

		   // amplitudes should be in time progression
		   //
		   if ( (((MyPoint)ukf.elementAt(i)).x == 
			 ((MyPoint)ukf.elementAt(j)).x) || 
			(((MyPoint)ukf.elementAt(i)).x > 
			 ((MyPoint)ukf.elementAt(j)).x) ) {
		       return false;
		   }
		}
	    }
	return true;
	}
    }

    /**
     *
     * Implementation of the run function from the Runnable interface.
     * Determines what the current step is and calls the appropriate method.
     *
     */
    public void run() {

	if (step_index_d == 1) {
	    disableControl();
	    step1();
	    enableControl();
 	}
	if (step_index_d == 2) {
	    disableControl();
	    step2();
	    enableControl(); 
	}
	if (step_index_d == 3) {
	    disableControl();
	    step3(); 
	    enableControl(); 
	}
	if (step_index_d == 4) {
	    disableControl();
	    step4(); 
	    enableControl(); 
	}
	if (step_index_d == 5) {
	    disableControl();
	    step5(); 
	    enableControl(); 
	}

	// exit gracefully
	//
	return;
    }
    
    /**
     *
     * Displays data sets from input box in output box, clears the data
     * points from the vector for input data, zero-mean data,
     * does interpolation, and then finds mean of the interpolated data
     * interpolates the zero-mean data, displays the result
     *
     * @return   True
     *
     */
    boolean step1() {

	// set up progress bar
	//
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);

    	// Display original data
	// 
	interpolated_set_1_d.removeAllElements();
	iset_1_d.removeAllElements();
	mset_1_d.removeAllElements();
	yestimate_1_d.removeAllElements();
	state_estimate_1_d.removeAllElements();
	interpol(set_1_d, interpolated_set_1_d);
	avg_1_d = mean(interpolated_set_1_d, mset_1_d);
	interpol(mset_1_d, iset_1_d);
	output_panel_d.addOutput(set_1_d, (Classify.PTYPE_INPUT), 
				 data_points_d.color_dset1);

	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	return true;
    }

   /**
    * find the LP coefficients using levinson durbin algorithm
    *
    * @return    true
    */
    boolean step2() {
	autoCorrelation();
	lpcCoefficient();

	double covar_nse_pro = covar_nse_err_d.getValue(lporder, lporder);
	double covar_nse_obs = covar_nse_err_d.getValue(lporder + 1, 
							lporder + 1);
	double covar_nse_err = covar_nse_err_d.getValue(0, 0);
	pro_box_d.appendMessage(new String(
			     " Class 0: \n\n" +
			     "    Variances \n" +
			     "     Q = process noise      = " +
			     MathUtil.SetDecimal(covar_nse_pro, 6) + "\n" +
			     "     R = observation noise  = " + 
			     MathUtil.SetDecimal(covar_nse_obs, 6) + "\n" +
			     "     P = estimate error     = " + 
			     MathUtil.SetDecimal(covar_nse_err, 6) + "\n"));
	return true;
    }   

   /**
    * Used to display and compute the first iteration of set_1_d
    *
    * @return    true
    *
    */
    boolean step3() {

	double[] wts_c = new double[sigma_len];
	double[] wts_m = new double[sigma_len];

	wts_m[0] = lambda_d / (aug_len + lambda_d);
	wts_c[0] = wts_m[0] + ( 1 - alpha_d * alpha_d + beta_d);

	for (int i = 1; i < sigma_len; i++) { 
	    wts_m[i] = 1.0 / (2 * (aug_len + lambda_d));
	    wts_c[i] = wts_m[i];
	}

 		int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(4);

	scale = iset_1_d.size();
	obs_seq_d = new double[scale];
	filt_data_d = new double[aug_len];

	for (int i = 0; i < scale; i++) {
	    obs_seq_d[i] = iset_1_d.elementAt(i).y;
	}

	state_mem_d.initMatrixValue((aug_len), 2 * (aug_len) + 1, 
				    0.0, Matrix.FULL);

	for ( ; idx < lporder; idx++) {
	    MyPoint p = new MyPoint();
	    p.y = ((MyPoint)iset_1_d.elementAt(idx)).y + avg_1_d;
	    p.x = ((MyPoint)iset_1_d.elementAt(idx)).x;
	    yestimate_1_d.addElement(new MyPoint(p));
	    
	    filt_data_d[idx] = obs_seq_d[idx];
	}
	filt_data_d[lporder] = 0.0;
	filt_data_d[lporder + 1] = 0.0;

	for (int i = 0; i < lporder; i++) {
	    double value = -finallpc_1_d[lporder - i];
	    trn_d.setValue(lporder - 1, i, value);
	}

	// square root of aug_len + lambda
	// 
	double sqrt_L_lambda = Math.sqrt(aug_len + lambda_d);

	double est_obsn = 0;
	Matrix temp = new Matrix();
	covar_nse_err_d.scalarMultMatrix(sqrt_L_lambda);	    
	temp.copyMatrix(covar_nse_err_d);
	covar_nse_err_d.sqrtMatrix(temp);
	covar_nse_err_d.copyMatrix(temp);
	    
	// first row will the previous state value
	// last two rows will be zero, as process and obsn
	// noise is assumed zero mean gaussian noise
	//
	for (int j = 0; j < sigma_len; j++) {
	    state_mem_d.addToColumn(filt_data_d, aug_len, j);
	}
	    
	// calculate the sigma points 
	// other column entries are sigma points hence, will have
	// either covar_nse_err_d column added or subtracted
	//
	for ( int j = 1; j < aug_len + 1; j++) {
	    double[] arr = new double[aug_len];
	    covar_nse_err_d.getColumn(arr, aug_len, (j-1));
	    state_mem_d.addToColumn(arr, aug_len, j);
	    state_mem_d.subFromColumn(arr, aug_len,(aug_len + j));
	}

	// time update - getting state sigma updates
	//
	Matrix state_dummy = new Matrix();
	state_dummy.initMatrixValue(lporder, sigma_len, 0.0, Matrix.FULL);
	state_mem_d.copyRows(state_dummy, 0, lporder);

	Matrix covar_nse_pro = new Matrix();
	covar_nse_pro.initMatrixValue(lporder, sigma_len, 
				      0.0, Matrix.FULL);
	double[] arr = new double[sigma_len];
	state_mem_d.getRow(arr, sigma_len, lporder);
	covar_nse_pro.addToRow(arr, sigma_len, lporder - 1);

	for (int j = 0; j < sigma_len; j++) {
	    Matrix state_coldata = new Matrix();
	    state_coldata.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	    Matrix pro_coldata = new Matrix();
	    pro_coldata.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	    state_dummy.copyColumns(state_coldata, j, 1);
	    covar_nse_pro.copyColumns(pro_coldata, j, 1);

	    Matrix mat = new Matrix();
	    trn_d.multMatrix(state_coldata, mat);
	    mat.addMatrix(pro_coldata);
	    for ( int k = 0; k < lporder; k++) {
		state_dummy.setValue(k, j, mat.getValue(k, 0));
	    }
	}

	// computing a mean state value based on weights w_m
	// and state sigmas
	//
	for (int k = 0; k < lporder; k++) {
	    filt_data_d[k] = 0.0;
	    for (int j = 0; j < sigma_len; j++) {
		double value = state_dummy.getValue(k, j);
		filt_data_d[k] = filt_data_d[k] + wts_m[j] * value;
	    }
	}

	// computing mean covar_nse_pro
	// based on wts_c, filt_data_d, state_dummy
	//
	Matrix subState = new Matrix();
	subState.initMatrixValue(lporder, sigma_len, 0.0, Matrix.FULL);

	Matrix subState_T = new Matrix();
	subState_T.initMatrixValue(sigma_len, lporder, 0.0, Matrix.FULL);

	Matrix state_covar = new Matrix();
	state_covar.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	for (int k = 0; k < lporder; k++) {
	    for (int j = 0; j < sigma_len; j++) {
		double val = state_dummy.getValue(k, j) - filt_data_d[k];
		subState.setValue(k, j, val);
	    }
	}
	subState.transposeMatrix(subState_T);

	for (int k = 0; k < lporder; k++) 
	    for (int j = 0; j < sigma_len; j++) 
		subState.setValue(k, j, 
				  wts_c[j] * subState.getValue(k, j));

	subState.multMatrix(subState_T, state_covar);

	// compute estimated observation sigmas and then the mean
	// of the estimated obsn sigmas
	//
	double[] y_est = new double[sigma_len];
	for (int j = 0; j < sigma_len; j++) {
	    y_est[j] = state_dummy.getValue(lporder - 1, j) + 
		state_mem_d.getValue(lporder + 1, j);
	    est_obsn = est_obsn + y_est[j] * wts_m[j];
	}

	// measurement update equations block
	// obsn covariance matrix
	//
	double covar_obs = 0.0;
	for ( int j = 0; j < sigma_len; j++) {
	    covar_obs = covar_obs + wts_c[j] *
		(y_est[j] - est_obsn)* (y_est[j] - est_obsn);
	}

	double[] cross_covar = new double[lporder];
	for(int k = 0; k < lporder; k++) {
	    cross_covar[k] = 0.0;
	    for ( int j = 0; j < sigma_len; j++) {
		cross_covar[k] = cross_covar[k] + 
		    wts_c[j] 
		    * (state_dummy.getValue(k, j) - filt_data_d[k]) 
		    * (y_est[j] - est_obsn);
	    }
	}

	// finding the kalman gain
	//
	Matrix kg = new Matrix();
	kg.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	for ( int j = 0; j < lporder; j++) {
	    double val = cross_covar[j] / covar_obs;
	    kg.setValue(j, 0, val);
	}

	// computing measurement update state equation
	//
	for (int j = 0; j < lporder; j++) {
	    filt_data_d[j] = filt_data_d[j] + kg.getValue(j, 0) * 
		(obs_seq_d[idx] - est_obsn);
	}

	// measurement update of state covar
	//
	Matrix mat = new Matrix();
	mat.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	Matrix mat_res = new Matrix();
	mat_res.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	Matrix kg_T = new Matrix();
	kg_T.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
	kg.transposeMatrix(kg_T);

	kg.multMatrix(kg_T, mat);
	mat.scalarMultMatrix(covar_obs);
	state_covar.subtractMatrix(mat, mat_res);
	state_covar.copyMatrix(mat_res);

	// realign the covar_nse_err_d matrix 
	//
	double covar_pro = covar_nse_err_d.getValue(lporder, lporder);
	covar_obs = covar_nse_err_d.getValue(lporder + 1, 
					     lporder + 1);

	covar_nse_err_d.initMatrixValue(lporder + 2, lporder + 2, 0.0,
					Matrix.FULL);

	for (int k = 0; k < lporder; k++) 
	    for (int j = 0; j < lporder; j++) 
		covar_nse_err_d.setValue(k, j, state_covar.getValue(k, j));
	covar_nse_err_d.setValue(lporder, lporder, covar_pro);
	covar_nse_err_d.setValue(lporder + 1, lporder + 1, covar_obs);

	// filtered data
	// 
	MyPoint p = new MyPoint();
	p.y = ((MyPoint)iset_1_d.elementAt(idx)).y;
	p.x = ((MyPoint)iset_1_d.elementAt(idx)).x;
	yestimate_1_d.addElement(new MyPoint(p));

	yestimate_1_d.elementAt(idx).y = 
	    filt_data_d[lporder - 1] + avg_1_d ;
	
// 	System.out.println("observation working on : " + idx);
// 	System.out.println(" actual obsn : " +
// 			   MathUtil.SetDecimal(obs_seq_d[idx], 6));
// 	System.out.println("filtered value : " 
// 			   + MathUtil.SetDecimal(val, 6));
	idx++;

	output_panel_d.addOutput(yestimate_1_d,
				 Classify.PTYPE_LINE,
				 Color.BLACK);
	output_panel_d.addOutput(yestimate_1_d,
				 (Classify.PTYPE_INPUT) * 4,
				 Color.BLACK);
	output_panel_d.repaint();
		
	try { Thread.sleep(50);
	}
	catch (Exception e) {}
		
	// removes the last output 
	//
	output_panel_d.removeOutput();
	output_panel_d.removeOutput();
	pro_box_d.setProgressCurr(++progress_curr);

	pro_box_d.setProgressCurr(4);
	return true;
    }   

   /**
    * Used to display and compute the second iteration of set_1_d
    *
    * @return    true
    *
    */
    boolean step4() {

	double[] wts_c = new double[sigma_len];
	double[] wts_m = new double[sigma_len];

	wts_m[0] = lambda_d / (aug_len + lambda_d);
	wts_c[0] = wts_m[0] + ( 1 - alpha_d * alpha_d + beta_d);

	for (int i = 1; i < sigma_len; i++) { 
	    wts_m[i] = 1.0 / (2 * (aug_len + lambda_d));
	    wts_c[i] = wts_m[i];
	}

 	
	int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(4);

	// square root of aug_len + lambda
	// 
	double sqrt_L_lambda = Math.sqrt(aug_len + lambda_d);

	double est_obsn = 0;
	Matrix temp = new Matrix();
	covar_nse_err_d.scalarMultMatrix(sqrt_L_lambda);	    
	temp.copyMatrix(covar_nse_err_d);
	covar_nse_err_d.sqrtMatrix(temp);
	covar_nse_err_d.copyMatrix(temp);
	    
	// first row will the previous state value
	// last two rows will be zero, as process and obsn
	// noise is assumed zero mean gaussian noise
	//
	for (int j = 0; j < sigma_len; j++) {
	    state_mem_d.addToColumn(filt_data_d, aug_len, j);
	}
	    
	// calculate the sigma points 
	// other column entries are sigma points hence, will have
	// either covar_nse_err_d column added or subtracted
	//
	for ( int j = 1; j < aug_len + 1; j++) {
	    double[] arr = new double[aug_len];
	    covar_nse_err_d.getColumn(arr, aug_len, (j-1));
	    state_mem_d.addToColumn(arr, aug_len, j);
	    state_mem_d.subFromColumn(arr, aug_len,(aug_len + j));
	}

	// time update - getting state sigma updates
	//
	Matrix state_dummy = new Matrix();
	state_dummy.initMatrixValue(lporder, sigma_len, 0.0, Matrix.FULL);
	state_mem_d.copyRows(state_dummy, 0, lporder);

	Matrix covar_nse_pro = new Matrix();
	covar_nse_pro.initMatrixValue(lporder, sigma_len, 
				      0.0, Matrix.FULL);
	double[] arr = new double[sigma_len];
	state_mem_d.getRow(arr, sigma_len, lporder);
	covar_nse_pro.addToRow(arr, sigma_len, lporder - 1);

	for (int j = 0; j < sigma_len; j++) {
	    Matrix state_coldata = new Matrix();
	    state_coldata.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	    Matrix pro_coldata = new Matrix();
	    pro_coldata.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	    state_dummy.copyColumns(state_coldata, j, 1);
	    covar_nse_pro.copyColumns(pro_coldata, j, 1);

	    Matrix mat = new Matrix();
	    trn_d.multMatrix(state_coldata, mat);
	    mat.addMatrix(pro_coldata);
	    for ( int k = 0; k < lporder; k++) {
		state_dummy.setValue(k, j, mat.getValue(k, 0));
	    }
	}

	// computing a mean state value based on weights w_m
	// and state sigmas
	//
	for (int k = 0; k < lporder; k++) {
	    filt_data_d[k] = 0.0;
	    for (int j = 0; j < sigma_len; j++) {
		double value = state_dummy.getValue(k, j);
		filt_data_d[k] = filt_data_d[k] + wts_m[j] * value;
	    }
	}

	// computing mean covar_nse_pro
	// based on wts_c, filt_data_d, state_dummy
	//
	Matrix subState = new Matrix();
	subState.initMatrixValue(lporder, sigma_len, 0.0, Matrix.FULL);

	Matrix subState_T = new Matrix();
	subState_T.initMatrixValue(sigma_len, lporder, 0.0, Matrix.FULL);

	Matrix state_covar = new Matrix();
	state_covar.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	for (int k = 0; k < lporder; k++) {
	    for (int j = 0; j < sigma_len; j++) {
		double val = state_dummy.getValue(k, j) - filt_data_d[k];
		subState.setValue(k, j, val);
	    }
	}
	subState.transposeMatrix(subState_T);

	for (int k = 0; k < lporder; k++) 
	    for (int j = 0; j < sigma_len; j++) 
		subState.setValue(k, j, 
				  wts_c[j] * subState.getValue(k, j));

	subState.multMatrix(subState_T, state_covar);

	// compute estimated observation sigmas and then the mean
	// of the estimated obsn sigmas
	//
	double[] y_est = new double[sigma_len];
	for (int j = 0; j < sigma_len; j++) {
	    y_est[j] = state_dummy.getValue(lporder - 1, j) + 
		state_mem_d.getValue(lporder + 1, j);
	    est_obsn = est_obsn + y_est[j] * wts_m[j];
	}

	// measurement update equations block
	// obsn covariance matrix
	//
	double covar_obs = 0.0;
	for ( int j = 0; j < sigma_len; j++) {
	    covar_obs = covar_obs + wts_c[j] *
		(y_est[j] - est_obsn)* (y_est[j] - est_obsn);
	}

	double[] cross_covar = new double[lporder];
	for(int k = 0; k < lporder; k++) {
	    cross_covar[k] = 0.0;
	    for ( int j = 0; j < sigma_len; j++) {
		cross_covar[k] = cross_covar[k] + 
		    wts_c[j] 
		    * (state_dummy.getValue(k, j) - filt_data_d[k]) 
		    * (y_est[j] - est_obsn);
	    }
	}

	// finding the kalman gain
	//
	Matrix kg = new Matrix();
	kg.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	for ( int j = 0; j < lporder; j++) {
	    double val = cross_covar[j] / covar_obs;
	    kg.setValue(j, 0, val);
	}

	// computing measurement update state equation
	//
	for (int j = 0; j < lporder; j++) {
	    filt_data_d[j] = filt_data_d[j] + kg.getValue(j, 0) * 
		(obs_seq_d[idx] - est_obsn);
	}

	// measurement update of state covar
	//
	Matrix mat = new Matrix();
	mat.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	Matrix mat_res = new Matrix();
	mat_res.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	Matrix kg_T = new Matrix();
	kg_T.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
	kg.transposeMatrix(kg_T);

	kg.multMatrix(kg_T, mat);
	mat.scalarMultMatrix(covar_obs);
	state_covar.subtractMatrix(mat, mat_res);
	state_covar.copyMatrix(mat_res);

	// realign the covar_nse_err_d matrix 
	//
	double covar_pro = covar_nse_err_d.getValue(lporder, lporder);
	covar_obs = covar_nse_err_d.getValue(lporder + 1, 
					     lporder + 1);

	covar_nse_err_d.initMatrixValue(lporder + 2, lporder + 2, 0.0,
					Matrix.FULL);

	for (int k = 0; k < lporder; k++) 
	    for (int j = 0; j < lporder; j++) 
		covar_nse_err_d.setValue(k, j, state_covar.getValue(k, j));
	covar_nse_err_d.setValue(lporder, lporder, covar_pro);
	covar_nse_err_d.setValue(lporder + 1, lporder + 1, covar_obs);

	// filtered data
	// 
	MyPoint p = new MyPoint();
	p.y = ((MyPoint)iset_1_d.elementAt(idx)).y;
	p.x = ((MyPoint)iset_1_d.elementAt(idx)).x;
	yestimate_1_d.addElement(new MyPoint(p));

	yestimate_1_d.elementAt(idx).y = 
	    filt_data_d[lporder - 1] + avg_1_d ;
	
// 	System.out.println("observation working on : " + idx);
// 	System.out.println(" actual obsn : " +
// 			   MathUtil.SetDecimal(obs_seq_d[idx], 6));
// 	System.out.println("filtered value : " 
// 			   + MathUtil.SetDecimal(val, 6));
	idx++;

	output_panel_d.addOutput(yestimate_1_d,
				 Classify.PTYPE_LINE,
				 Color.BLACK);
	output_panel_d.addOutput(yestimate_1_d,
				 (Classify.PTYPE_INPUT) * 4,
				 Color.BLACK);
	output_panel_d.repaint();
		
	try { Thread.sleep(50);
	}
	catch (Exception e) {}
		
	// removes the last output 
	//
	output_panel_d.removeOutput();
	output_panel_d.removeOutput();
	pro_box_d.setProgressCurr(++progress_curr);

	pro_box_d.setProgressCurr(4);
	return true;
    }   
 
    /**
     * @return   true
     * 
     */
    boolean step5() {

	double[] wts_c = new double[sigma_len];
	double[] wts_m = new double[sigma_len];

	wts_m[0] = lambda_d / (aug_len + lambda_d);
	wts_c[0] = wts_m[0] + ( 1 - alpha_d * alpha_d + beta_d);

	for (int i = 1; i < sigma_len; i++) { 
	    wts_m[i] = 1.0 / (2 * (aug_len + lambda_d));
	    wts_c[i] = wts_m[i];
	}

 	
	int progress_curr = 0;
	pro_box_d.setProgressCurr(0);
	pro_box_d.setProgressMax(scale);

	// square root of aug_len + lambda
	// 
	double sqrt_L_lambda = Math.sqrt(aug_len + lambda_d);

	for (int i = 0; i < scale - lporder - 2; i++) {
	// for (int i = 0; i < 1; i++) {

	    double est_obsn = 0;
	    Matrix temp = new Matrix();
	    covar_nse_err_d.scalarMultMatrix(sqrt_L_lambda);	    
	    temp.copyMatrix(covar_nse_err_d);
	    covar_nse_err_d.sqrtMatrix(temp);
	    covar_nse_err_d.copyMatrix(temp);
	    
	    // first row will the previous state value
	    // last two rows will be zero, as process and obsn
	    // noise is assumed zero mean gaussian noise
	    //
	    for (int j = 0; j < sigma_len; j++) {
		state_mem_d.addToColumn(filt_data_d, aug_len, j);
	    }
	    
	    // calculate the sigma points 
	    // other column entries are sigma points hence, will have
	    // either covar_nse_err_d column added or subtracted
	    //
	    for ( int j = 1; j < aug_len + 1; j++) {
		double[] arr = new double[aug_len];
		covar_nse_err_d.getColumn(arr, aug_len, (j-1));
		state_mem_d.addToColumn(arr, aug_len, j);
		state_mem_d.subFromColumn(arr, aug_len,(aug_len + j));
	    }

	    // time update - getting state sigma updates
	    //
	    Matrix state_dummy = new Matrix();
	    state_dummy.initMatrixValue(lporder, sigma_len, 0.0, Matrix.FULL);
	    state_mem_d.copyRows(state_dummy, 0, lporder);

	    Matrix covar_nse_pro = new Matrix();
	    covar_nse_pro.initMatrixValue(lporder, sigma_len, 
					  0.0, Matrix.FULL);
	    double[] arr = new double[sigma_len];
	    state_mem_d.getRow(arr, sigma_len, lporder);
	    covar_nse_pro.addToRow(arr, sigma_len, lporder - 1);

	    for (int j = 0; j < sigma_len; j++) {
		Matrix state_coldata = new Matrix();
		state_coldata.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		Matrix pro_coldata = new Matrix();
		pro_coldata.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
		state_dummy.copyColumns(state_coldata, j, 1);
		covar_nse_pro.copyColumns(pro_coldata, j, 1);

		Matrix mat = new Matrix();
		trn_d.multMatrix(state_coldata, mat);
		mat.addMatrix(pro_coldata);
		for ( int k = 0; k < lporder; k++) {
		    state_dummy.setValue(k, j, mat.getValue(k, 0));
		}
	    }

	    // computing a mean state value based on weights w_m
	    // and state sigmas
	    //
	    for (int k = 0; k < lporder; k++) {
		filt_data_d[k] = 0.0;
		for (int j = 0; j < sigma_len; j++) {
		    double value = state_dummy.getValue(k, j);
		    filt_data_d[k] = filt_data_d[k] + wts_m[j] * value;
		}
	    }

	    // computing mean covar_nse_pro
	    // based on wts_c, filt_data_d, state_dummy
	    //
	    Matrix subState = new Matrix();
	    subState.initMatrixValue(lporder, sigma_len, 0.0, Matrix.FULL);

	    Matrix subState_T = new Matrix();
	    subState_T.initMatrixValue(sigma_len, lporder, 0.0, Matrix.FULL);

	    Matrix state_covar = new Matrix();
	    state_covar.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	    for (int k = 0; k < lporder; k++) {
		for (int j = 0; j < sigma_len; j++) {
		    double val = state_dummy.getValue(k, j) - filt_data_d[k];
		    subState.setValue(k, j, val);
		}
	    }
	    subState.transposeMatrix(subState_T);

	    for (int k = 0; k < lporder; k++) 
		for (int j = 0; j < sigma_len; j++) 
		    subState.setValue(k, j, 
				      wts_c[j] * subState.getValue(k, j));

	    subState.multMatrix(subState_T, state_covar);

	    // compute estimated observation sigmas and then the mean
	    // of the estimated obsn sigmas
	    //
	    double[] y_est = new double[sigma_len];
	    for (int j = 0; j < sigma_len; j++) {
		y_est[j] = state_dummy.getValue(lporder - 1, j) + 
		    state_mem_d.getValue(lporder + 1, j);
		est_obsn = est_obsn + y_est[j] * wts_m[j];
	    }

	    // measurement update equations block
	    // obsn covariance matrix
	    //
	    double covar_obs = 0.0;
	    for ( int j = 0; j < sigma_len; j++) {
		covar_obs = covar_obs + wts_c[j] *
		    (y_est[j] - est_obsn)* (y_est[j] - est_obsn);
	    }

	    double[] cross_covar = new double[lporder];
	    for(int k = 0; k < lporder; k++) {
		cross_covar[k] = 0.0;
		for ( int j = 0; j < sigma_len; j++) {
		    cross_covar[k] = cross_covar[k] + 
			wts_c[j] 
			* (state_dummy.getValue(k, j) - filt_data_d[k]) 
			* (y_est[j] - est_obsn);
		}
	    }

	    // finding the kalman gain
	    //
	    Matrix kg = new Matrix();
	    kg.initMatrixValue(lporder, 1, 0.0, Matrix.FULL);
	    for ( int j = 0; j < lporder; j++) {
		double val = cross_covar[j] / covar_obs;
		kg.setValue(j, 0, val);
	    }

	    // computing measurement update state equation
	    //
	    for (int j = 0; j < lporder; j++) {
		filt_data_d[j] = filt_data_d[j] + kg.getValue(j, 0) * 
		    (obs_seq_d[idx] - est_obsn);
	    }

	    // measurement update of state covar
	    //
	    Matrix mat = new Matrix();
	    mat.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);
	    Matrix mat_res = new Matrix();
	    mat_res.initMatrixValue(lporder, lporder, 0.0, Matrix.FULL);

	    Matrix kg_T = new Matrix();
	    kg_T.initMatrixValue(1, lporder, 0.0, Matrix.FULL);
	    kg.transposeMatrix(kg_T);

	    kg.multMatrix(kg_T, mat);
	    mat.scalarMultMatrix(covar_obs);
	    state_covar.subtractMatrix(mat, mat_res);
	    state_covar.copyMatrix(mat_res);

	    // realign the covar_nse_err_d matrix 
	    //
	    double covar_pro = covar_nse_err_d.getValue(lporder, lporder);
	    covar_obs = covar_nse_err_d.getValue(lporder + 1, 
							    lporder + 1);

	    covar_nse_err_d.initMatrixValue(lporder + 2, lporder + 2, 0.0,
					    Matrix.FULL);

	    for (int k = 0; k < lporder; k++) 
		for (int j = 0; j < lporder; j++) 
		    covar_nse_err_d.setValue(k, j, state_covar.getValue(k, j));
	    covar_nse_err_d.setValue(lporder, lporder, covar_pro);
	    covar_nse_err_d.setValue(lporder + 1, lporder + 1, covar_obs);

	    // filtered data
	    // 
	    MyPoint p = new MyPoint();
	    p.y = ((MyPoint)iset_1_d.elementAt(idx)).y;
	    p.x = ((MyPoint)iset_1_d.elementAt(idx)).x;
	    yestimate_1_d.addElement(new MyPoint(p));

	    yestimate_1_d.elementAt(idx).y = 
		filt_data_d[lporder - 1] + avg_1_d ;
	    
// 	    System.out.println("observation working on : " + idx);
// 	    System.out.println(" actual obsn : " +
// 			       MathUtil.SetDecimal(obs_seq_d[idx], 6));
// 	    System.out.println("filtered value : " 
// 			       + MathUtil.SetDecimal(val, 6));
	    idx++;

	    output_panel_d.addOutput(yestimate_1_d,
				     Classify.PTYPE_LINE,
				     Color.BLACK);
	    output_panel_d.addOutput(yestimate_1_d,
				     (Classify.PTYPE_INPUT) * 4,
				     Color.BLACK);
	    output_panel_d.repaint();
		
	    try { Thread.sleep(50);
	    }
	    catch (Exception e) {}
		
	    // removes the last output 
	    //
	    output_panel_d.removeOutput();
	    pro_box_d.setProgressCurr(++progress_curr);
 	}


       pro_box_d.setProgressCurr(scale);

       double compute_err = 
	   actual_error( yestimate_1_d, interpolated_set_1_d);

        pro_box_d.appendMessage(" Class 0: \n" + 
				"   Mean Square Error   = " +
				 MathUtil.SetDecimal(compute_err, 6) + "\n");
 	pro_box_d.appendMessages("Algorithm Complete \n");
	pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);
	output_panel_d.repaint();

       	return true;
    }
 
    /**
     *
     * Calculates the interpolated points for the data inputs   
     *
     * @param  v  input data points
     * @param  iset  interpolated data points
     *
     */
    public void interpol(Vector<MyPoint> v, Vector<MyPoint> iset)
    {	
	double sample;
	double d ;
	MyPoint r = new MyPoint();

	double[] y2 = new double[v.size()];  
	double[] y  = new double[v.size()];
	double[] x  = new double[v.size()];
       
       for (int k = 0; k < v.size(); k++)
       {
	   x[k] = ((MyPoint)v.elementAt(k)).x;
	   y[k] = ((MyPoint)v.elementAt(k)).y;
       }

       // Reference:
       // "Numerical Recipe in C", Cambridge University Press, pp 113-116
       //
       // total time interval of the data points calculated
       //
       d = ((MyPoint)v.lastElement()).x - ((MyPoint)v.firstElement()).x ; 
       
       // time step calculation
       //
       sample = d / scale;
       spline(x, y, y2, v.size());
       
       // iset holds new data points - interpolated values
       // First value added 
       //
       iset.addElement(new MyPoint((MyPoint)v.firstElement()));
       
       // Increment to next Time value
       //
      	r.x = sample + ((MyPoint)v.firstElement()).x;

	while (r.x <=  ((MyPoint)v.lastElement()).x)
	{
	    for(int j = 1; j < v.size(); j++)
	    {
		// the New Interpolated amplitude will be calculated
		// for the two points within which the time intervals 
		// lies, Hence, the condition check
		//
		if((r.x <= ((MyPoint)v.elementAt(j)).x ) && 
		   (r.x >=  ((MyPoint)v.elementAt(j - 1)).x)) 
		{
		    r.y = 0 ;
		    splint ((MyPoint)v.elementAt(j - 1), 
			    (MyPoint)v.elementAt(j), r, y2, (j - 1));

		    // New value appended to the array
		    //
		    iset.addElement(new MyPoint(r));
		    break; 
		}
	    }	    
   	    r.x = r.x + sample;
	}
	// sometimes the last element of the user entered 
	// values is missed out hence the condition to 
	// cross-check and put it in
	//
	if (((MyPoint)iset.lastElement()).x != ((MyPoint)v.lastElement()).x )
	    iset.addElement( new MyPoint((MyPoint)v.lastElement()));
    }

    /**
     *
     * Actually interpolates the points
     *
     * @param x    array containing the x coordinates of datapoints
     * @param y    array containing the y coordinates of datapoints
     * @param y2   array containing the interpolated y coordinates
     * @param size the size of the array to be interpolated
     * 
     */
    public void spline (double[] x, double[] y, double[] y2, int size )
    { 
	double[] u = new double[size];
	double p; 
	double sig; 
	
	y2[0] = 0;
	u[0] = 0;
	 
	for( int i = 1; i < size - 1; i++ )
	{ 
	    sig = ( x[i] - x[i - 1] ) / ( x[i + 1] - x[i - 1] ) ;
	    p = sig * y2[i - 1] + 2 ;
	    y2[i] = ( sig - 1) / p ;
	    u[i] = ( y[i + 1] - y[i] ) / ( x[i + 1] - x[i] ) ;
	    u[i] = ( 6 * u[i] / ( x[i + 1] - x[i - 1] ) 
		     - sig * u[i - 1] ) / p ;
	}
	y2[size - 1] = 0 ;
	
	// This is the back substitution loop of the tridiagonal algorithm
	//
	for(int i = size - 2; i >= 0; i-- )     
	    y2[i] = y2[i] * y2[i + 1] + u[i] ;
    }

    /**
     * Interpolates for a point between the two known points
     * using Cubic Interpolation
     *
     * @param   u1 start point for the interpolation
     * @param   u2 end point for the interpolation
     * @param   r  returning point, basically the interpolated point
     * @param   y2 array used for reassigning of r
     * @param   i  the sample number
     *
     */
    public void splint ( MyPoint u1, MyPoint u2, MyPoint r, 
			 double[] y2, int i )
    {
	double h;
	double a;
	double b;
 
	h = u2.x - u1.x ;
	a = ( u2.x - r.x ) / h ;
	b = ( r.x - u1.x ) / h ;
	r.y = a * u1.y + b * u2.y + ((a * a * a - a) * y2[i] 
	                             + ( b * b * b - b ) * y2[i + 1] ) 
	                                                 * ( h * h ) / 6 ; 
    }

    /**
     *
     * Calculates the mean and the zero-mean data points  
     *
     * @param    v orginal datapoints
     * @param    mv zero mean datapoints 
     *
     * @return   The average in a double
     * 
     */
    public double mean(Vector<MyPoint> v, Vector<MyPoint> mv)
    {
        double average = 0;
	MyPoint p      = new MyPoint();

       for(int i = 0; i < v.size(); i++ )
	   average = average + ((MyPoint)v.elementAt(i)).y;
       average = average / v.size();
       
       for(int i = 0; i < v.size(); i++ )
       {
	   p.y = ((MyPoint)v.elementAt(i)).y - average;
	   p.x =  ((MyPoint)v.elementAt(i)).x; 
	   mv.addElement(new MyPoint(p));
       }        
       return average ;
    }  
   
    /**
    *
    * Computes the autocorrelation coeffient from the data sets   
    *
    */
    public void autoCorrelation()
    {
	if (iset_1_d.size() > 0 )
	{
	    autoc_1_d = new double[lporder + 1];
	    autocorrelate(iset_1_d, autoc_1_d);
	}	
    }

    /**
     *
     * Actaully computes the autocorrelation coefficients
     *
     * @param  v Vector of datapoints
     * @param  autoCoeff_co array of autocorrelation coefficients
     *
     */
    public void autocorrelate ( Vector v, double[] autoCoeff_co) 
    {
	int size;
	size                 = v.size();
      	double[] amplitude_y = new double[size];

      	// to store the amplitude for calculation autocoefficient
      	//
    	for (int i = 0; i < size; i++)
	    amplitude_y[i] = ((MyPoint)v.elementAt(i)).y;

    	for (int j = 0; j <= lporder; j++)
      	{
      	    autoCoeff_co[j] = 0;
      	    if ( j < size)
      		for (int k = 0; k < size - j; k++)
		    autoCoeff_co[j] = autoCoeff_co[j] 
			            + amplitude_y[k] * amplitude_y[k + j];
	}

   	// normalize the autocorrelation cofficient
      	//
     	for (int i = lporder; i >= 0; i--) 
      	    autoCoeff_co[i] = autoCoeff_co[i] / autoCoeff_co[0];
    }

    /**
     *
     * Computes the Linear Prediction coefficient from the data sets   
     *
     */
    public void lpcCoefficient()
    {
	// Rabiner, Schafer, "Digital Processing of Speech signals", 
	// Bell Laboratories, Inc. 1978, 
	// section 8.3.2 Durbin's recursive solution for 
	// the autocorrelation equations pp 411-412
	//
	if (set_1_d.size() > 0)
	{
	    finallpc_1_d = new double [lporder + 1];
	    ref_coeff_1_d = new double [lporder + 1];
	    est_err_1_d = calculate_lpc(autoc_1_d, 
					    finallpc_1_d, ref_coeff_1_d);
	}
    }

    /**
     *
     * Actually calculate the LP coefficient and the Residual Error 
     * Energy, and Reflection Coefficients
     *
     * @param  auto_coeff array of auto correlation coefficients
     * @param  lpc        array of linear prediction coefficients
     * @param  rc_reg     array of reflection coefficients
     * 
     * @return residual error energy in a double
     *
     */
    public double calculate_lpc (double[] auto_coeff, 
				 double[] lpc, double[] rc_reg)
    {
	int j_bckwd       = 0;
	long middle_index = 0;

	// Reference:
	// J.D. Markel and A.H. Gray, "Linear Prediction of Speech,"
	// Springer-Verlag Berlin Heidelberg, New York, USA, pp. 219,
	// 1980.
 
	// initialization
	//
	lpc[0] = 1;
	double err_energy = auto_coeff[0];

	// if the energy of the signal is zero we set all the predictor
	// coefficients to zero and exit
	//
	if (err_energy == (double)0)
	{
	    for (int i = 0; i < lporder + 1 ; i++)
		lpc[i] = 0;
	}
	else
        {
	    // do the first step manually
	    //
	    double sum = 0;
	    double tmp = 0;
	    
	    sum         = -auto_coeff[1] / err_energy;
	    rc_reg[1]   = sum;
	    lpc[1]      = rc_reg[1];
	    tmp         = 1 - rc_reg[1] * rc_reg[1];
	    err_energy *= tmp;
	    // recursion
	    //
	    for (int i = 2; i <= lporder; i++)
	    {
		sum = 0;
		for (int j = 1; j < i; j++)
		    sum += lpc[j] * auto_coeff[i - j];
 
		rc_reg[i]    = -(auto_coeff[i] + sum) / err_energy;	    
		lpc[i]       = rc_reg[i];
		j_bckwd      = i - 1;
		middle_index = i / 2;
		for (int j = 1; j <= middle_index; j++)
		{
		    sum          = lpc[j_bckwd] + rc_reg[i] * lpc[j];
		    lpc[j]       = lpc[j] + rc_reg[i] * lpc[i - j];
		    lpc[j_bckwd] = sum;
		    j_bckwd--;
		}
		// compute new error
		//
		tmp = 1.0 - rc_reg[i] * rc_reg[i];
		err_energy *= tmp;
	    }
	}
	return err_energy;
    }

    /**
     *
     * Compute the actual error from the given data points and the estimated
     * values.
     *
     * @param  y_estimate datapoints of the estimated datapoints
     * @param  iset       original datapoints 
     *
     * @return actual error energy in a double
     */
    public double actual_error (Vector y_estimate, Vector iset)
    {
	double error_value;
	double act_error = (double) 0;
	int j = 0;
	int i = 0;

	// first point and the last point have their x-coordinates same
	// So the difference in y-values is the error value
	//
	error_value = ((MyPoint)y_estimate.firstElement()).y 
	              - ((MyPoint)iset.firstElement()).y;
	act_error = act_error + error_value * error_value; 
	error_value = ((MyPoint)y_estimate.lastElement()).y 
	              - ((MyPoint)iset.lastElement()).y;
	act_error = act_error + error_value * error_value;

	// for next values
	// need to continue till all the user defined points are exhausted
	//
	j++;

	for (i = 1; ( (i < y_estimate.size()) && (j < iset.size()) ); i++)
	{ 
	    while ( (((MyPoint)y_estimate.elementAt(i)).x < 
		     ((MyPoint)iset.elementAt(j)).x) 
		    && ((i < y_estimate.size() - 1) && (j < iset.size())) )
	    {
		    i++;
	    }

	    if ( j < iset.size())
	    {
		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     == ((MyPoint)iset.elementAt(j)).x ) 
		{
		    error_value = ((MyPoint)y_estimate.elementAt(i)).y 
			          - ((MyPoint)iset.elementAt(j)).y;
		    act_error = act_error + error_value * error_value;
		}

		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     > ((MyPoint)iset.elementAt(j)).x )
		{
		    double y1 = ((MyPoint)y_estimate.elementAt(i)).y;
		    double y2 = ((MyPoint)y_estimate.elementAt(i - 1)).y;
		    double x1 = ((MyPoint)y_estimate.elementAt(i)).x;
		    double x2 = ((MyPoint)y_estimate.elementAt(i - 1)).x;
		    double x_unknown = ((MyPoint)iset.elementAt(j)).x;
		    error_value = y2 - ( (x2 - x_unknown) * 
					 ( y2 - y1) / (x2 - x1));
		    act_error = act_error + error_value * error_value ;
		}
	    }
	    j++;
	}
	return act_error;
    }

} //end class
